/*

trcommon.c

Authors:
        Tatu Ylonen <ylo@ssh.fi>
        Markku-Juhani Saarinen <mjos@ssh.fi>
        Timo J. Rinne <tri@ssh.fi>
        Sami Lehtinen <sjl@ssh.fi>

Copyright (C) 1996-2002 SSH Communications Security Corp, Helsinki, Finland
All rights reserved.

*/
/*
 TODO:

 - create a mapping between algorithm names in the draft and crypto library
 
*/

#include "sshincludes.h"
#include "sshbufaux.h"
#include "sshgetput.h"
#include "sshsnlist.h"
#include "sshcipherlist.h"
#include "sshtimeouts.h"
#include "sshtcp.h"
#include "sshtrans.h"
#include "sshmsgs.h"
#include "sshencode.h"
#include "trcommon.h"
#include "trkex.h"
#include "ssh2pubkeyencode.h"
#include "sshregex.h"
#include "sshdsprintf.h"

#define SSH_DEBUG_MODULE "Ssh2Transport"

void ssh_tr_process_output(SshTransportCommon tr);
void ssh_tr_process_input(SshTransportCommon tr);
void ssh_tr_key_check_done_make_kex2_cb(SshBuffer packet,
                                        SshTransportCommon tr);
SshBuffer ssh_tr_input_packet(SshTransportCommon tr);
unsigned int ssh_tr_peek_packet_type(SshBuffer packet);

/* Uncomment this to dump packet contents. */
/* #define DUMP_PACKETS */

/* Performs cleanups after a key exchange.  This can also be used to
   prepare for a new key exchange. */

void ssh_tr_kex_cleanup(SshTransportCommon tr)
{
  SSH_DEBUG(4, ("Cleaning up after a key exchange."));

  if (tr->client_kexinit_packet)
    {
      ssh_buffer_free(tr->client_kexinit_packet);
      tr->client_kexinit_packet = NULL;
    }
  if (tr->server_kexinit_packet)
    {
      ssh_buffer_free(tr->server_kexinit_packet);
      tr->server_kexinit_packet = NULL;
    }
  if (tr->client_kex1_packet)
    {
      ssh_buffer_free(tr->client_kex1_packet);
      tr->client_kex1_packet = NULL;
    }
  if (tr->server_kex1_packet)
    {
      ssh_buffer_free(tr->server_kex1_packet);
      tr->server_kex1_packet = NULL;
    }
  if (tr->server)
    {
      /* XXX */
    }
  else
    {
      if (tr->public_host_key_blob)
        {
          ssh_buffer_free(tr->public_host_key_blob);
          tr->public_host_key_blob = NULL;
        }
      if (tr->public_host_key)
        {
          ssh_public_key_free(tr->public_host_key);
          tr->public_host_key = NULL;
        }
    }
}

/* Destroys our context immediately.  This means that we will release all
   resources now. */

void ssh_tr_destroy_now(SshTransportCommon tr)
{
  SSH_TRACE(3, ("Destroying SshTransportCommon object."));
  if (tr == NULL)
    return;
  ssh_cancel_timeouts(SSH_ALL_CALLBACKS, (void *)tr);
  ssh_tr_kex_cleanup(tr);
  if (tr->connection)
    ssh_stream_destroy(tr->connection);
  tr->connection = NULL;
  if (tr->params)
    ssh_transport_destroy_params(tr->params);
  ssh_buffer_uninit(&tr->outgoing);
  if (tr->incoming_packet)
    ssh_buffer_free(tr->incoming_packet);
  SSH_ASSERT(tr->up_stream == NULL);  /* Should be... */
  tr->up_callback = NULL_FNPTR;
  ssh_buffer_uninit(&tr->up_outgoing);
  ssh_buffer_uninit(&tr->up_incoming);
  ssh_xfree(tr->own_version);
  ssh_xfree(tr->guessed_kex);
  ssh_xfree(tr->guessed_host_key);
  ssh_xfree(tr->kex_name);
  ssh_xfree(tr->host_key_name);
  ssh_xfree(tr->host_key_names);
  ssh_xfree(tr->c_to_s.cipher_name);
  ssh_xfree(tr->c_to_s.mac_name);
  ssh_xfree(tr->c_to_s.compression_name);
  ssh_xfree(tr->s_to_c.cipher_name);
  ssh_xfree(tr->s_to_c.mac_name);
  ssh_xfree(tr->s_to_c.compression_name);
  if (tr->hash)
    ssh_hash_free(tr->hash);
  if (tr->outgoing_cipher)
    ssh_cipher_free(tr->outgoing_cipher);
  if (tr->incoming_cipher)
    ssh_cipher_free(tr->incoming_cipher);
  if (tr->outgoing_mac)
    ssh_mac_free(tr->outgoing_mac);
  if (tr->incoming_mac)
    ssh_mac_free(tr->incoming_mac);
  if (tr->compression_outgoing)
    ssh_compress_free(tr->compression_outgoing);
  if (tr->compression_incoming)
    ssh_compress_free(tr->compression_incoming);
  if (tr->compression_buffer)
    ssh_buffer_free(tr->compression_buffer);
  ssh_xfree(tr->service_name);
  if (tr->server_host_name)
    ssh_xfree(tr->server_host_name);

  if (tr->public_host_key)
    ssh_public_key_free(tr->public_host_key);
  
  memset(tr->session_identifier, 0, tr->session_identifier_len);
  memset(tr->exchange_hash, 0, tr->exchange_hash_len);

  /* clear and free dh key exchange stuff */

  ssh_mp_clear(tr->dh_p);
  ssh_mp_clear(tr->dh_g);
  ssh_mp_clear(tr->dh_e);
  ssh_mp_clear(tr->dh_f);
  ssh_mp_clear(tr->dh_k);
  ssh_mp_clear(tr->dh_secret);

  /* Fill with garbage for debugging. */
  memset(tr, 'F', sizeof(*tr));
  ssh_xfree(tr);
}

unsigned int tr_get_padding_byte(SshTransportCommon tr)
{
  size_t padding_pool_len;
  Boolean just_initialized = FALSE;
  int i;
  unsigned int byte;

#if (defined CIPHER_PADDING_GENERATOR) && !(defined PADDING_GENERATOR)
#define PADDING_GENERATOR "aes"
#endif

  padding_pool_len = sizeof(tr->padding_pool);
  if (tr->padding_gen_c == 0)
    {
      /* Initialization: this happens either when (1) this function is
         called for the first time after initialization of tr or (2)
         when the counter tr->padding_gen_c reaches 0 through
         incrementation.  */

      SSH_TRACE(6, ("Keying padding pool generator."));
      just_initialized = TRUE;
      
      {
#ifndef CIPHER_PADDING_GENERATOR  /* use weak pseudo-randomness from
                                     sshcore/sshrand */
        SshUInt32 seed;
        for (i = 0; i < sizeof(seed); i++)
          *(((unsigned char*)&seed) + i) = ssh_random_get_byte();

        SSH_DEBUG_HEXDUMP(9, ("Seed:"), (unsigned char *)&seed, sizeof(seed));
        ssh_rand_state_seed(&tr->prng_state, seed);
        seed = 0L;
#else /* CIPHER_PADDING_GENERATOR */ /* use cipher (default: AES) */
        unsigned char key[16];

        for (i = 0; i < sizeof(key); i++)
          key[i] = ssh_random_get_byte();
      
        if (tr->padding_gen)
          ssh_cipher_free(tr->padding_gen);

        if (ssh_cipher_allocate(PADDING_GENERATOR, key, sizeof(key), TRUE,
                                &tr->padding_gen) != SSH_CRYPTO_OK)
          ssh_fatal("Failed to initialize %s for calculating padding bytes.",
                    PADDING_GENERATOR);

        memset(key, 0, sizeof(key));
#endif /* CIPHER_PADDING_GENERATOR */
      }
    }

  if (tr->padding_pool_c < padding_pool_len && !just_initialized)
    {
      byte = tr->padding_pool[tr->padding_pool_c];
    }
  else
    {
      SSH_DEBUG(7, ("Running padding generator."));
      tr->padding_pool_c = 0;
#ifndef CIPHER_PADDING_GENERATOR  /* use weak pseudo-randomness from
                                     sshcore/sshrand */
      for (i = 0; i < 4; i++)
        SSH_PUT_32BIT(&tr->padding_pool[i*4], ssh_rand_state(&tr->prng_state));
#else /* CIPHER_PADDING_GENERATOR */ /* use cipher (default: AES). */

      /* (the pool is initialized with the padding_gen_c counter, and
         then encrypted in cbc mode.)  */
      memset(tr->padding_pool, 0, padding_pool_len);
      SSH_PUT_32BIT(tr->padding_pool, tr->padding_gen_c);
      if (ssh_cipher_transform(tr->padding_gen,
                               tr->padding_pool,
                               tr->padding_pool,
                               padding_pool_len) != SSH_CRYPTO_OK)
        ssh_fatal("Failed to generate more bytes to the padding pool.");
#endif /* CIPHER_PADDING_GENERATOR */
      SSH_DEBUG_HEXDUMP(10, ("Pool:"), tr->padding_pool, padding_pool_len);
      byte = tr->padding_pool[0];
      tr->padding_gen_c++;
    }
  tr->padding_pool_c++;
  return byte;
}

/* Called from the bottom of the event loop, we try to process input from
   the connection.  This will restart automatic reading. */

void ssh_tr_process_input_proc(void *context)
{
  SshTransportCommon tr = context;

  ssh_tr_process_input(tr);
}

/* Ensure that we will proceed reading, even if reading was previously
   blocked. */

void ssh_tr_wake_up_input(SshTransportCommon tr)
{
  /* If reads not blocked, just return. */
  if (!tr->read_has_blocked)
    return;

  SSH_DEBUG(9, ("Waking blocked reads."));

  tr->read_has_blocked = FALSE;
  ssh_register_timeout(0L, 0L, ssh_tr_process_input_proc, (void *)tr);
}

/* Call the application callback.  This is called from the event loop only. */

void ssh_tr_up_signal_input_proc(void *context)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(9, ("Calling application callback for reads."));

  if (tr->up_callback)
    (*tr->up_callback)(SSH_STREAM_INPUT_AVAILABLE, tr->up_context);
}

/* Call the application callback.  This is called from the event loop only. */

void ssh_tr_up_signal_output_proc(void *context)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(9, ("Calling application callback for writes."));

  if (tr->up_callback)
    (*tr->up_callback)(SSH_STREAM_CAN_OUTPUT, tr->up_context);
}

/* Signal that the application can read. */

void ssh_tr_up_signal_input(SshTransportCommon tr)
{
  if (tr->up_callback)
    ssh_register_timeout(0L, 0L, ssh_tr_up_signal_input_proc, (void *)tr);
}

/* Signal that the application can write. */

void ssh_tr_up_signal_output(SshTransportCommon tr)
{
  if (tr->up_callback)
    ssh_register_timeout(0L, 0L, ssh_tr_up_signal_output_proc, (void *)tr);
}

/* Sends a packet upstream.  Buffers the given data, and signals a wakeup
   for the stream if appropriate. */

void ssh_tr_up_send(SshTransportCommon tr, SshCrossPacketType type,
                    const unsigned char *payload, size_t len)
{
  SSH_DEBUG(8, ("Sending packet up. Type: %d, len: %ld", type, len));
  /* Wake up reads from up if they have blocked. */
  if (tr->up_read_blocked)
    { 
      ssh_tr_up_signal_input(tr);
      tr->up_read_blocked = FALSE;
    }
  
  /* Add data to the upwards outgoing buffer. */
  ssh_bufaux_put_int(&tr->up_outgoing, (len + 1));
  ssh_bufaux_put_char(&tr->up_outgoing, (int)type);
  ssh_xbuffer_append(&tr->up_outgoing, payload, len);
}

/* Send as much data as possible from the outgoing buffer.  Return TRUE
   if all data was sent and processing can continue, and FALSE if an
   error occurred or we can do no more. */

Boolean ssh_tr_output_outgoing(SshTransportCommon tr)
{
  int len;
#ifdef HAVE_SIGNAL
  void (*old_handler)(int);
  
  /* Note that this is ugly, but we must do this to get a proper
     disconnect message in case upper level has not ignored SIGPIPE.  I
     did some performance testing, and this doesn't seem to have a
     noticeable effect on performance. */
  old_handler = signal(SIGPIPE, SIG_IGN);
#endif /* HAVE_SIGNAL */

  while (ssh_buffer_len(&tr->outgoing) > 0)
    {
      len = ssh_buffer_len(&tr->outgoing);
      SSH_DEBUG(9, ("Writing %ld bytes.", len));

      len = ssh_stream_write(tr->connection,
                             ssh_buffer_ptr(&tr->outgoing),
                             len);
      if (len < 0)
        {
          SSH_DEBUG(9, ("Cannot write more at this time. "));
#ifdef HAVE_SIGNAL
          signal(SIGPIPE, old_handler);
#endif /* HAVE_SIGNAL */
          return FALSE;    /* We cannot write more at this time. */
        }
      if (len == 0)
        {
          SshBuffer disconnect_packet;
          /* We cannot write any more; presumably the connection has been
             lost. */

          if ((disconnect_packet = ssh_tr_input_packet(tr)) != NULL)
            {
              /* If a complete packet is received after write has
                 permanently failed, it might be a proper disconnect
                 packet. If it was a disconnect packet, it would've been
                 handled by ssh_tr_input_packet(). */
              SSH_DEBUG(2, ("Received packet type %d after write has "
                            "permanently failed (expecting "
                            "SSH_MSG_DISCONNECT).",
                            ssh_tr_peek_packet_type(disconnect_packet)));
              ssh_buffer_free(disconnect_packet);
            }
          else
            {
              /* If we did received a disconnect packet,
                 ssh_tr_up_disconnect() will check whether it has
                 already disconnected, so there is no harm in calling it
                 multiple times. */
              ssh_tr_up_disconnect(tr, TRUE, FALSE,
                                   SSH_DISCONNECT_CONNECTION_LOST,
                                   "Connection lost on output.");
            }
          ssh_buffer_clear(&tr->outgoing);
#ifdef HAVE_SIGNAL
          signal(SIGPIPE, old_handler);
#endif /* HAVE_SIGNAL */
          return FALSE;
        }
      ssh_buffer_consume(&tr->outgoing, len);
    }

#ifdef HAVE_SIGNAL
  signal(SIGPIPE, old_handler);
#endif /* HAVE_SIGNAL */

  /* Send an eof to the connection if requested. */
  if (tr->outgoing_eof)
    ssh_stream_output_eof(tr->connection);

  /* Check if we should destroy the context next */
  if (tr->up_stream == NULL)
    {
      ssh_tr_destroy_now(tr);
      return FALSE;
    }

  /* Wake up writes from up if enough space in buffer. */
  if (tr->up_write_blocked &&
      ssh_buffer_len(&tr->up_outgoing) <
      XMALLOC_MAX_SIZE - SSH_MAX_TOTAL_PACKET_LENGTH - SSH_CONTROL_RESERVE &&
      ssh_buffer_len(&tr->outgoing) < SSH_BUFFERING_LIMIT)
    {
      SSH_DEBUG(9, ("Waking up application output."));
      tr->up_write_blocked = FALSE;
      ssh_tr_up_signal_output(tr);
    }

  SSH_DEBUG(9, ("No more data to write."));
  return TRUE;
}

/* Wraps the packet structure around the payload in the buffer.*/
void ssh_tr_encode_packet(SshTransportCommon tr,
                          const unsigned char *payload,
                          size_t payload_length)
{
  size_t block_size, length, padding_length, mac_length;
  int i;
  unsigned char *start;
  unsigned char seq_buf[4];

  SSH_DEBUG(9, ("Sending packet with type %d to connection", payload[0]));

  if (tr->outgoing_eof)
    {
      /* Trying to send after we have sent EOF??? */
      SSH_TRACE(0, ("trying to send after EOF."));
      return;
    }

  /* Compress the payload if appropriate. */
  tr->uncompressed_outgoing_bytes += payload_length;
  if (!ssh_compress_is_none(tr->compression_outgoing))
    {
      ssh_buffer_clear(tr->compression_buffer);
      ssh_compress_buffer(tr->compression_outgoing, payload, payload_length,
                          tr->compression_buffer);
      payload = ssh_buffer_ptr(tr->compression_buffer);
      payload_length = ssh_buffer_len(tr->compression_buffer);
    }
  tr->compressed_outgoing_bytes += payload_length;

  /* Compute restrictions for encryption block size. */
  block_size = ssh_cipher_get_block_length(tr->outgoing_cipher);
  if (block_size < 8)
    block_size = 8;

  mac_length = ssh_mac_length(tr->outgoing_mac);

  /* Compute padding length and the total length */

  length = 1 + 4 + payload_length;
  padding_length = block_size - (length % block_size);
  if (padding_length < 4)
    padding_length += block_size;

  length += padding_length;  /* now everything but the mac */

  SSH_DEBUG(9, ("length %d pad %d payload %d mac %d",
                length, padding_length, payload_length, mac_length));

  /* Store the plaintext packet in the buffer. */

  ssh_xbuffer_append_space(&tr->outgoing, &start, length + mac_length);
  SSH_PUT_32BIT(start, length - 4);  /* not including the length itself */
  start[4] = padding_length;
  memcpy(4 + 1 + start, payload, payload_length);

  for (i = 0; i < padding_length; i++)
    start[4 + 1 + payload_length + i] = tr_get_padding_byte(tr);

  /* Compute and store MAC. */

  if (!tr->ssh_old_mac_bug_compat)
    {
      /* Everything is a-ok */
      ssh_mac_start(tr->outgoing_mac);
      SSH_PUT_32BIT(seq_buf, tr->outgoing_sequence_number);
      ssh_mac_update(tr->outgoing_mac, seq_buf, 4);
      ssh_mac_update(tr->outgoing_mac, start, length);
      ssh_mac_final(tr->outgoing_mac, start + length);
    }
  else
    {
      /* other side counts the MAC in the old (==wrong) way. */
      ssh_mac_start(tr->outgoing_mac);
      ssh_mac_update(tr->outgoing_mac, start, length);
      SSH_PUT_32BIT(seq_buf, tr->outgoing_sequence_number);
      ssh_mac_update(tr->outgoing_mac, seq_buf, 4);
      ssh_mac_final(tr->outgoing_mac, start + length);
    }

#ifdef DUMP_PACKETS
  SSH_DEBUG_HEXDUMP(3, ("Outgoing packet:"),
                    ssh_buffer_ptr(&tr->outgoing),
                    ssh_buffer_len(&tr->outgoing));
#endif /* DUMP_PACKETS */

  /* Encrypt the packet (but not the MAC). */
  if (ssh_cipher_transform(tr->outgoing_cipher, start, start, length) !=
      SSH_CRYPTO_OK)
    /* XXX Probably should disconnect. */
    ssh_fatal("ssh_tr_encode_packet: encrypting failed (length %d)", length);

  /* Increment the packet sequence number. */
  tr->outgoing_sequence_number++;

#ifdef DUMP_PACKETS
  SSH_DEBUG_HEXDUMP(3, ("Outgoing packet -- encrypted --:"),
                    ssh_buffer_ptr(&tr->outgoing),
                    ssh_buffer_len(&tr->outgoing));
#endif /* DUMP_PACKETS */
}

/* Wraps the packet structure around the payload in the buffer,
   and sends it out. */

void ssh_tr_send_packet(SshTransportCommon tr,
                        const unsigned char *payload,
                        size_t payload_length)
{
  if (ssh_buffer_len(&tr->outgoing) <= 0)
    {
      SshBufferStruct new_packet;
      size_t ret;
      ssh_buffer_init(&new_packet);
      /* Sending empty ignore packet to counter attack against "
         "CBC-mode ciphers. */
      SSH_DEBUG(5, ("Outgoing empty, sending empty ignore packet."));
      ret = ssh_encode_buffer(&new_packet,
                              SSH_FORMAT_CHAR, (unsigned int) SSH_MSG_IGNORE,
                              SSH_FORMAT_UINT32, 0L,
                              SSH_FORMAT_END);
      SSH_VERIFY(ret > 0);
      ssh_tr_encode_packet(tr, ssh_buffer_ptr(&new_packet),
                           ssh_buffer_len(&new_packet));
      ssh_buffer_uninit(&new_packet);
    }
  ssh_tr_encode_packet(tr, payload, payload_length);
  /* Start writing if not already active. */
  ssh_tr_output_outgoing(tr);
}

/* Terminates the protocol, and sends a disconnect message up. */

void ssh_tr_up_disconnect(SshTransportCommon tr,
                          Boolean locally_generated,
                          Boolean send_to_other_side,
                          unsigned int reason,
                          const char *fmt, ...)
{
  SshBufferStruct buffer;
  char message[512];
  va_list va;

  if (tr->up_outgoing_eof)
    {
      /* We are already disconnected - just return. */
      SSH_TRACE(2, ("Already disconnected."));
      return;
    }

  /* Format the message. */
  va_start(va, fmt);
  ssh_vsnprintf(message, sizeof(message), fmt, va);
  va_end(va);

  SSH_DEBUG(3, ("Disconnecting: reason code: %d message: '%.200s'",
                (int)reason, message));

  /* If appropriate, send the disconnect to the other side. */
  if (send_to_other_side)
    {
      ssh_buffer_init(&buffer);
      ssh_bufaux_put_char(&buffer, SSH_MSG_DISCONNECT);
      ssh_bufaux_put_int(&buffer, reason);
      ssh_bufaux_put_uint32_string(&buffer, message, strlen(message));
      ssh_bufaux_put_uint32_string(&buffer, "en", 2);
      ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
      ssh_buffer_uninit(&buffer);
    }

  /* Format the upstream packet payload. */
  ssh_buffer_init(&buffer);
  ssh_bufaux_put_boolean(&buffer, locally_generated);
  ssh_bufaux_put_int(&buffer, reason);
  ssh_bufaux_put_uint32_string(&buffer, message, strlen(message));
  ssh_bufaux_put_uint32_string(&buffer, "en", 2);

  /* Send and free the packet. */
  ssh_tr_up_send(tr, SSH_CROSS_DISCONNECT,
                 ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);

  /* Prepare for shutdown. */
  tr->up_outgoing_eof = TRUE;
  tr->received_state = RECEIVED_DEAD;
  tr->sent_state = SENT_DEAD;
  tr->outgoing_eof = TRUE;
  ssh_tr_output_outgoing(tr);
}

/* Read in version number.  This returns TRUE if more input may be
   available and reading should continue. */

Boolean ssh_tr_input_version(SshTransportCommon tr)
{
  int len;
  char *verstring_p, *temp_verstring;
  static int string_index = 0;
  /* XXX This should be done a bit smarter, as we MAY display the
     information sent by the server.*/
  static char temp_string[100];

  SSH_DEBUG(5, ("Reading version number."));

  /* According to the draft: The server MAY send other lines of
     data before sending the version string.  Each line SHOULD be
     terminated by a carriage return and newline.  Such lines MUST NOT
     begin with "SSH-"... */

  /* check if we havent received 'SSH-' and we are client */
  if (tr->remote_version_index == 0 && !tr->server)
    {
      for (;;)
        {
          /* Read a single character.  We cannot read more as we are
             waiting 'SSH-' */
          len = ssh_stream_read(tr->connection,
                                (unsigned char *)temp_string +
                                string_index, 1);
          if (len == 0)
            {
              ssh_tr_up_disconnect(tr, TRUE, FALSE,
                                   SSH_DISCONNECT_CONNECTION_LOST,
                                   "Connection closed by remote host.");
              return FALSE;
            }
          if (len < 0)
            return FALSE; /* No more data available yet. */

          if (string_index < 3)
            {
              string_index++;
              continue;
            }

          /* Check if we have a match */
          if (!memcmp(&temp_string[string_index - 3], "SSH-", 4))
            {
              memmove(tr->remote_version, "SSH-", 4);
              tr->remote_version_index = 4;
              break;
            }

          /* Don't read it too much. Wrap around. (It would be
             against the draft to disconnect. Theoretically, the
             server should be allowed to send the collected works
             of Shakespeare if it wanted to.) */
          if (string_index >= sizeof(temp_string) - 1)
            {
              /* Move the last 4 read chars to the beginning, so we
                 don't lose any part of the (possible) version string */
              memmove(temp_string, &temp_string[string_index - 3], 4);
              string_index = 4;
              continue;
            }

          /* Count these characters. */
          string_index++;
        }
    }

  /* Keep reading until the version identifier has been received. */
  for (;;)
    {
      /* Is it too long? */
      if (tr->remote_version_index == sizeof(tr->remote_version) - 1)
        {
          ssh_tr_up_disconnect(tr, TRUE, FALSE,
                               SSH_DISCONNECT_PROTOCOL_ERROR,
                               "Remote protocol version string too long.");
          return FALSE;
        }

      /* Read a single character.  We cannot read more as we are waiting
         for a newline. */
      len = ssh_stream_read(tr->connection, (unsigned char *)
                            tr->remote_version + tr->remote_version_index, 1);
      if (len == 0)
        {
          ssh_tr_up_disconnect(tr, TRUE, FALSE,
                               SSH_DISCONNECT_CONNECTION_LOST,
                               "Connection closed by remote host.");
          return FALSE;
        }
      if (len < 0)
        return FALSE; /* No more data available yet. */

      /* Check if we are at end of version id. Note that we don't include the
         newline in the version number string. */
      if (tr->remote_version[tr->remote_version_index] == '\n')
        break;

      if (tr->remote_version[tr->remote_version_index] == '\r')
        continue;  /* Ignore carriage return. */

      /* Count these characters. */
      tr->remote_version_index++;
    }

  /* Null-terminate the version number. */
  tr->remote_version[tr->remote_version_index] = '\0';

  /* Show the remote version in debug output. */
  ssh_debug("Remote version: %s", tr->remote_version);

  /* Call the version callback, so that it can check for ssh1 compatibility. */
  if (tr->version_callback)
    {
      SshStream conn_tmp = tr->connection;
      /* If version callback calls ssh_conn_destroy immediately,
         it has already shut down the connection stream.  So here
         tr->connection is temporarily set to NULL and restored,
         if version callback does not take over the stream. */
      tr->connection = NULL;
      if ((*tr->version_callback)(tr->remote_version, tr->version_context) ==
          FALSE)
        {
          /* The compatibility code took over the protocol conversation.
             We need to disarm the transport object and return.
             No traffic is sent through this level any more. */
          tr->connection = NULL;
#if 0
          /* This is mystically called through ssh_conn_destroy
             after the emulation library has closed the session. */
          tr->up_stream = NULL;
          ssh_tr_destroy_now(tr);
#endif
          return FALSE; /*XXX*/
        }
      else
        {
          tr->connection = conn_tmp;
        }
    }

  /* This version of ssh will only talk to a SSH 2.0 host. */
  /* XXX forwards compatibility with 3.0 (or whatever) We should be able
     to handle that (remember 1.2.25?) */
  if (strncmp(tr->remote_version, "SSH-2.", 6) != 0 &&
      strncmp(tr->remote_version, "SSH-1.99", 8) != 0)
    {
      ssh_tr_up_disconnect(tr, TRUE, FALSE,
                           SSH_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED,
                           "Illegal protocol version.");

      return FALSE;
    }

  /* Compatibility with older ssh-2.0.x versions */

  if ((verstring_p = strchr(tr->remote_version, '-')) != NULL)
    {
      verstring_p++;
      if ((verstring_p = strchr(verstring_p, '-')) != NULL)
        {
          SshRegexContext rctx;
          SshRegexMatcher rmatcher;
          char *major_version_str, *minor_version_str, *revision_str;
          int major_version = 0, minor_version = 0, revision = 0;

          verstring_p++;
          temp_verstring = ssh_xstrdup(verstring_p);
          if ((verstring_p = strchr(temp_verstring, ' ')) != NULL)
            {
              verstring_p = '\0';
            }

          /* Decode the version string */
          rctx = ssh_regex_create_context();
          rmatcher =
            ssh_regex_create(rctx, "^(~d+)~.(~d+){~.(~d+)}?.*",
                             SSH_REGEX_SYNTAX_SSH);

          if (ssh_regex_match_cstr(rmatcher, temp_verstring))
            {
              major_version_str = (char *)ssh_regex_get_submatch(rmatcher, 1);
              minor_version_str = (char *)ssh_regex_get_submatch(rmatcher, 2);
              revision_str = (char *)ssh_regex_get_submatch(rmatcher, 3);

              major_version = atoi(major_version_str);
              minor_version = atoi(minor_version_str);
              if (revision_str)
                revision = atoi(revision_str);
              else
                revision = 0;
              
              ssh_debug("Major: %d Minor: %d Revision: %d",
                        major_version, minor_version, revision);

              if (major_version == 2 && minor_version == 0 && revision <= 9)
                {
                  SSH_DEBUG(2, ("Remote version has MAC calculation "
                                "order bug."));
                  tr->ssh_old_mac_bug_compat = TRUE;
                }
              if (major_version == 2 && minor_version == 0 && revision <= 10)
                {
                  SSH_DEBUG(2, ("Remote version has key size reduction bug."));
                  tr->ssh_old_keygen_bug_compat = TRUE;
                }
              if (major_version == 2 && minor_version == 0 && revision <= 12)
                {
                  SSH_DEBUG(2, ("Remote version has publickey draft "
                                "incompatibility bug."));
                  tr->ssh_old_publickey_bug_compat = TRUE;
                }
              if (major_version == 2 && minor_version == 0 && revision <= 13)
                {
                  SSH_DEBUG(2, ("Remote version has service accept message "
                                "draft incompatibility bug."));
                  tr->ssh_old_msg_service_accept_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has publickey service name "
                                "draft incompatibility bug."));
                  tr->ssh_old_publickey_service_name_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has X11 channel open "
                                "draft incompatibility bug."));
                  tr->ssh_old_x11_channel_open_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has "
                                "SSH_MSG_CHANNEL_OPEN_FAILURE "
                                "draft incompatibility bug."));
                  tr->ssh_old_channel_open_failure_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has "
                                "SSH_MSG_USERAUTH_PK_OK "
                                "draft incompatibility bug."));
                  tr->ssh_old_publickey_pk_ok_msg_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has rekey "
                                "incompatibility bug."));
                  tr->ssh_old_rekey_compat = TRUE;
                }

              if ((major_version == 2 && minor_version == 0) ||
                  (major_version == 2 && minor_version == 1 && revision == 0))
                {
                  SSH_DEBUG(2, ("Remote version has hostbased service name "
                                "draft incompatibility bug."));
                  tr->ssh_old_hostbased_service_name_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has publickey session_id "
                                "encoding draft incompatibility bug."));
                  tr->ssh_old_publickey_session_id_encoding_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has malformed signatures "
                                "draft incompatibility bug."));
                  tr->ssh_old_malformed_signatures_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version uses deprecated "
                                "disconnect codes."));
                  tr->ssh_old_deprecated_disconnect_codes_compat = TRUE;
                }

              if (major_version == 2 && minor_version <= 4)
                {
                  SSH_DEBUG(2, ("Remote version has kex packet guess "
                                "determination bug."));
                  tr->ssh_old_guess_determination_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version has hostbased looping on "
                                "failure bug."));
                  tr->ssh_old_hostbased_loop_bug_compat = TRUE;

                  SSH_DEBUG(2, ("Remote version can only handle one key in "
                                "\"hostbased\" auth."));
                  tr->ssh_old_hostbased_single_key_bug_compat = TRUE;

#ifdef SSHDIST_CRYPT_RSA
                  SSH_DEBUG(2, ("Remote version uses md5 instead of sha-1 as "
                                "hash algorithm with RSA keys."));
                  tr->ssh_old_rsa_hash_scheme_bug_compat = TRUE;
#endif /* SSHDIST_CRYPT_RSA */

                  if (minor_version <= 3 &&
                      /* F-Secure's 2.1.1, 2.2.1 and 2.3.1 versions behaved
                         correctly. */
                      !((minor_version == 1 || minor_version == 2 ||
                         minor_version == 3) && revision == 1))
                    {
                      SSH_DEBUG(2, ("Remote version uses 16 byte "
                                    "key with SHA-1."));
                      tr->ssh_old_sha1_key_len_bug_compat = TRUE;
                    }
                  if (!(minor_version == 4 && revision >= 2))
                    {
                      SSH_DEBUG(2, ("Remote version doesn't send language tag "
                                    "in SSH_MSG_USERAUTH_PASSWD_CHANGEREQ."));
                      tr->ssh_old_no_langtag_in_passwd_chreq_bug_compat = TRUE;
                    }
                }
              if ((major_version == 3 && minor_version <= 1) ||
                  (major_version == 2 && minor_version <= 4))
                {
                  SSH_DEBUG(2, ("Remote version calls \"des-cbc@ssh.com\" "
                                "\"des-cbc\"."));
                  tr->ssh_old_des_cbc_naming_bug_compat = TRUE;
                }
            }
          ssh_regex_free(rmatcher);

          /* Other manufacturers versions. */
          rmatcher =
            ssh_regex_create(rctx, "^OpenSSH_(~d+)~.(~d+){~.(~d+)}?.*",
                             SSH_REGEX_SYNTAX_SSH);
          if (ssh_regex_match_cstr(rmatcher, temp_verstring))
            {
              major_version_str = (char *)ssh_regex_get_submatch(rmatcher, 1);
              minor_version_str = (char *)ssh_regex_get_submatch(rmatcher, 2);
              revision_str = (char *)ssh_regex_get_submatch(rmatcher, 3);

              major_version = atoi(major_version_str);
              minor_version = atoi(minor_version_str);
              if (revision_str)
                revision = atoi(revision_str);
              else
                revision = 0;
              
              ssh_debug("OpenSSH: Major: %d Minor: %d Revision: %d",
                        major_version, minor_version, revision);

              SSH_DEBUG(2, ("All versions of OpenSSH handle kex guesses "
                            "incorrectly."));
              if (!tr->server)
                tr->send_kex1_immediately = FALSE;

              if (major_version < 3)
                {
                  SSH_DEBUG(2, ("Remote version has rekey incompatibility "
                                "bug."));
                  tr->ssh_old_rekey_compat = TRUE;
                }
            }

          ssh_regex_free(rmatcher);
          
          rmatcher =
            ssh_regex_create(rctx, "^Sun_SSH_(~d+)~.(~d+){~.(~d+)}?.*",
                             SSH_REGEX_SYNTAX_SSH);
          if (ssh_regex_match_cstr(rmatcher, temp_verstring))
            {
              major_version_str = (char *)ssh_regex_get_submatch(rmatcher, 1);
              minor_version_str = (char *)ssh_regex_get_submatch(rmatcher, 2);
              revision_str = (char *)ssh_regex_get_submatch(rmatcher, 3);

              major_version = atoi(major_version_str);
              minor_version = atoi(minor_version_str);
              if (revision_str)
                revision = atoi(revision_str);
              else
                revision = 0;

              ssh_debug("Sun_SSH: Major: %d Minor: %d Revision: %d",
                        major_version, minor_version, revision);

              if (major_version == 1 && minor_version < 1)
                {
                  /* Darren Moffat of Sun assured me that these will be
                     fixed by 1.1, so I'm putting my confidence on that. */
                  SSH_DEBUG(2, ("SunSSH_1.0 handles kex guesses "
                                "incorrectly."));
                  if (!tr->server)
                    tr->send_kex1_immediately = FALSE;

                  SSH_DEBUG(2, ("Remote version has rekey incompatibility "
                                "bug."));
                  tr->ssh_old_rekey_compat = TRUE;
                }
            }

          ssh_regex_free(rmatcher);
          ssh_regex_free_context(rctx);
          ssh_xfree(temp_verstring);
        }
    }

  /* We have now received the entire remote version number. */
  tr->received_state = RECEIVED_VERSION;

  /* Try output, as this may enable us to do something new. */
  ssh_tr_process_output(tr);

  return TRUE;
}

/* Constructs a KEX1 packet, if appropriate for the current kex method and
   algorithms.  Returns the packet payload if one is to be sent, or NULL
   if no such packet should be sent for the current method.  Saves the
   packet as our KEX1 packet.  It is legal to call this multiple times. */

SshBuffer ssh_tr_make_kex1(SshTransportCommon tr)
{
  SshBuffer packet;

  SSH_DEBUG(4, ("Constructing the first key exchange packet."));

  /* Create the KEX1 packet and save it.  Note that the returned packet may
     also be NULL. */
  if (tr->server)
    {
      packet = (*tr->kex->server_make_kex1)(tr);
      if (tr->server_kex1_packet)
        ssh_buffer_free(tr->server_kex1_packet);
      tr->server_kex1_packet = packet;
    }
  else
    {
      packet = (*tr->kex->client_make_kex1)(tr);
      if (tr->client_kex1_packet)
        ssh_buffer_free(tr->client_kex1_packet);
      tr->client_kex1_packet = packet;
    }

  return packet;
}

/* Constructs a KEX2 packet, if appropriate for the current kex method
   and algorithms.  Returns the packet payload by calling the
   completion procedure, or NULL if no KEX2 packet is to be sent for
   the current kex method.  The sent state must be updated by the
   caller. The buffer the packet is returned in must eventually be
   freed by the caller. */

void ssh_tr_make_kex2(SshTransportCommon tr,
                      SshMakeKex2CompletionProc completion)
{
  SSH_DEBUG(4, ("Constructing the second key exchange packet."));
 
  /* Create the KEX2 packet. Note that the returned packet may also be
     NULL. */
  if (tr->server)
    (*tr->kex->server_make_kex2)(tr, completion);
  else
    (*tr->kex->client_make_kex2)(tr, completion);
}

/* Get a string from both the client and the server key exchange
   packets, and choose an algorithm from those.  Returns the chosen
   algorithm (allocated with ssh_xmalloc; the caller must free it with
   ssh_xfree) or NULL if no common algorithm can be found. */

char *ssh_tr_negotiate_one_alg(SshTransportCommon tr,
                               const char *description,
                               SshBuffer client_kexinit,
                               SshBuffer server_kexinit)
{
  char *client_list, *server_list, *common_list, *result;

  if (ssh_decode_buffer(client_kexinit,
                        SSH_FORMAT_UINT32_STR, &client_list, NULL,
                        SSH_FORMAT_END) == 0)
    return NULL;
  if (ssh_decode_buffer(server_kexinit,
                        SSH_FORMAT_UINT32_STR, &server_list, NULL,
                        SSH_FORMAT_END) == 0)
    {
      ssh_xfree(client_list);
      return NULL;
    }
  common_list = ssh_snlist_intersection(client_list, server_list);

  if (common_list == NULL || strlen(common_list) == 0)
    {
      SSH_TRACE(2, ("Algorithm negotiation failed for %s: client list: %s vs. "
                    "server list : %s",
                    description, client_list, server_list));
      ssh_xfree(client_list);
      ssh_xfree(server_list);
      if (common_list)
        ssh_xfree(common_list);

      return NULL;
    }

  result = strchr(common_list, ',');
  if (result)
    {
      *result = '\0';
    }

  result = ssh_xstrdup(common_list);

  ssh_xfree(client_list);
  ssh_xfree(server_list);
  ssh_xfree(common_list);
  return result;
}

/* Assings a new value to the char pointer, freeing the old value if it is
   non-NULL.  The argument string must have been allocated by ssh_xmalloc;
   this does not copy it. */

void ssh_tr_set_string(char **cpp, char *new_value)
{
  if (*cpp != NULL)
    ssh_xfree(*cpp);
  *cpp = new_value;
}

/* Compute the algorithms that result from the negotiation.  This returns
   TRUE if a common set of algorithms was chosen, and FALSE if no compatible
   set of algorithms can be chosen.  If TRUE is returned, this sets
   guess_was_wrong to indicate whether the initial guessed algorithm was
   correct. */

Boolean ssh_tr_negotiate(SshTransportCommon tr,
                         Boolean *guess_was_wrong)
{
  char *client_kex, *server_kex, *common_kex, *kex;
  char *client_server_host_key, *server_server_host_key, *common_host_key;
  char *chosen_kex = NULL, *chosen_host_key = NULL;
  char *chosen_c_to_s_cipher = NULL, *chosen_s_to_c_cipher = NULL;
  char *chosen_c_to_s_mac = NULL, *chosen_s_to_c_mac = NULL;
  char *chosen_c_to_s_compression = NULL, *chosen_s_to_c_compression = NULL;
  char *client_preferred_kex, *server_preferred_kex,
    *client_preferred_host_key, *server_preferred_host_key;
  SshBuffer client_kexinit, server_kexinit;

  SSH_DEBUG(3, ("Computing algorithms from key exchange."));

  /* We must have both kexinit packets available. */
  SSH_ASSERT(tr->client_kexinit_packet != NULL);
  SSH_ASSERT(tr->server_kexinit_packet != NULL);

  /* Copy the kexinit packets into local buffers. */
  client_kexinit = ssh_xbuffer_allocate();
  ssh_xbuffer_append(client_kexinit, ssh_buffer_ptr(tr->client_kexinit_packet),
                     ssh_buffer_len(tr->client_kexinit_packet));
  server_kexinit = ssh_xbuffer_allocate();
  ssh_xbuffer_append(server_kexinit, ssh_buffer_ptr(tr->server_kexinit_packet),
                     ssh_buffer_len(tr->server_kexinit_packet));
  
  /* Parse relevant information from the client packet. */
  if (ssh_decode_buffer(client_kexinit,
                        SSH_FORMAT_CHAR, NULL,
                        SSH_FORMAT_DATA, NULL, 16,
                        SSH_FORMAT_UINT32_STR, &client_kex, NULL,
                        SSH_FORMAT_UINT32_STR, &client_server_host_key, NULL,
                        SSH_FORMAT_END) == 0)
    {
      SSH_TRACE(0, ("Error parsing client kexinit packet."));
      return FALSE;
    }

  /* Parse the same information from the server packet. */
  if (ssh_decode_buffer(server_kexinit,
                        SSH_FORMAT_CHAR, NULL,
                        SSH_FORMAT_DATA, NULL, 16,
                        SSH_FORMAT_UINT32_STR, &server_kex, NULL,
                        SSH_FORMAT_UINT32_STR, &server_server_host_key, NULL,
                        SSH_FORMAT_END) == 0)
    {
      SSH_TRACE(0, ("Error parsing server kexinit packet."));
      ssh_xfree(client_kex);
      ssh_xfree(client_server_host_key);
      return FALSE;
    }

  /* The kex method will have to be supported by both. */
  common_kex = ssh_snlist_intersection(client_kex, server_kex);
  /* Compute which host key types are supported by both. */
  common_host_key = ssh_snlist_intersection(client_server_host_key,
                                            server_server_host_key);

  SSH_TRACE(4, ("client: kex = %s, hk_alg = %s", client_kex,
                client_server_host_key));
  SSH_TRACE(4, ("server: kex = %s, hk_alg = %s", server_kex,
                server_server_host_key));

  /* Note the preferred algorithms of both sides */
  client_preferred_kex = ssh_snlist_get_name(client_kex);
  client_preferred_host_key = ssh_snlist_get_name(client_server_host_key);

  server_preferred_kex = ssh_snlist_get_name(server_kex);
  server_preferred_host_key = ssh_snlist_get_name(server_server_host_key);

  ssh_xfree(client_kex);
  ssh_xfree(server_kex);

  if (client_preferred_kex == NULL ||
      client_preferred_host_key == NULL ||
      server_preferred_kex == NULL ||
      server_preferred_host_key == NULL)
    {
      ssh_warning("ssh_tr_negotiate: failed to get kex or hostkey "
                  "algorithm names. (possibly a malformed kex-packet "
                  "(possible attack attempt))");
      return FALSE;
    }
  
  /* Guess is wrong if the preferred algorithms differ (guessed
     algorithm must be the preferred algorithm (the first one in the
     list)). This does flag some right guesses wrong in simpler kex
     algorithms (like diffie-hellman) but this is what the protocol
     spec says. As long as there are no inconsistencies or false
     positives, this is not a problem. */

  *guess_was_wrong = (strcmp(client_preferred_kex,
                             server_preferred_kex) != 0 ||
                      strcmp(client_preferred_host_key,
                             server_preferred_host_key) != 0);

  SSH_DEBUG(6, ("Host key algorithm list from client: %s",
                client_server_host_key));
  SSH_DEBUG(6, ("Host key algorithm list from server: %s",
                server_server_host_key));

  if (tr->host_key_names)
    ssh_xfree(tr->host_key_names);

  tr->host_key_names = ssh_xstrdup(common_host_key);

  ssh_xfree(client_server_host_key);
  ssh_xfree(server_server_host_key);

  /* Loop over the common kex methods. */
  chosen_host_key = NULL;

  kex = common_kex;

  while (kex)
    {
      char *host_key_copy = ssh_xstrdup(common_host_key);
      char *hk, *tmp_kex;

      hk = host_key_copy;
      while (hk)
        {
          char *tmp_hk;

          /* XXX if hk does not support signature/encryption as needed by kex,
             then continue. */
          if (1)
            {
              /* Terminate string on next delimeter, if one is found. */
              tmp_kex = strchr(kex, ',');
              if (tmp_kex)
                {
                  *tmp_kex = '\0';
                }

              /* Terminate string on next delimeter, if one is found. */
              tmp_hk = strchr(hk, ',');
              if (tmp_hk)
                {
                  *tmp_hk = '\0';
                }

              chosen_kex = ssh_xstrdup(kex);
              chosen_host_key = ssh_xstrdup(hk);
              break;
            }

          tmp_hk = strchr(hk, ',');
          if (tmp_hk)
            {
              /* Advance over the delimiter. */
              tmp_hk++;
            }
          hk = tmp_hk;
        }
      ssh_xfree(host_key_copy);
      if (chosen_host_key != NULL)
        break;

      tmp_kex = strchr(kex, ',');
      if (tmp_kex)
        {
          /* Advance over the delimiter. */
          tmp_kex++;
        }
      kex = tmp_kex;
    }

  ssh_xfree(common_kex);
  ssh_xfree(common_host_key);

  if (strlen(chosen_kex) == 0)
    {
      ssh_xfree(chosen_kex);
      chosen_kex = NULL;
    }
  if (strlen(chosen_host_key) == 0)
    {
      ssh_xfree(chosen_host_key);
      chosen_host_key = NULL;
    }

  /* Choose the remaining algorithms. */
  chosen_c_to_s_cipher = ssh_tr_negotiate_one_alg(tr, "c_to_s_cipher",
                                                  client_kexinit,
                                                  server_kexinit);
  chosen_s_to_c_cipher = ssh_tr_negotiate_one_alg(tr, "s_to_c_cipher",
                                                  client_kexinit,
                                                  server_kexinit);
  chosen_c_to_s_mac = ssh_tr_negotiate_one_alg(tr, "c_to_s_mac",
                                               client_kexinit,
                                               server_kexinit);
  chosen_s_to_c_mac = ssh_tr_negotiate_one_alg(tr, "s_to_c_mac",
                                               client_kexinit,
                                               server_kexinit);
  chosen_c_to_s_compression = ssh_tr_negotiate_one_alg(tr, "c_to_s_compr",
                                                       client_kexinit,
                                                       server_kexinit);
  chosen_s_to_c_compression = ssh_tr_negotiate_one_alg(tr, "s_to_c_compr",
                                                       client_kexinit,
                                                       server_kexinit);

  {
    char *lang_c_to_s, *lang_s_to_c;
    SshBuffer packet = tr->server ? client_kexinit : server_kexinit;

    lang_c_to_s = ssh_bufaux_get_uint32_string(packet, NULL);
    lang_s_to_c = ssh_bufaux_get_uint32_string(packet, NULL);

    SSH_TRACE(2, ("lang s to c: `%s', lang c to s: `%s'",
                  lang_c_to_s, lang_s_to_c));
    ssh_xfree(lang_s_to_c);
    ssh_xfree(lang_c_to_s);
  }

  if (tr->server)
    tr->first_kex_packet_follows = ssh_bufaux_get_boolean(client_kexinit);
  else
    tr->first_kex_packet_follows = ssh_bufaux_get_boolean(server_kexinit);

  SSH_DEBUG(4, ("first_kex_packet_follows: %s",
                tr->first_kex_packet_follows ? "TRUE" : "FALSE"));

  if (chosen_kex == NULL || chosen_host_key == NULL)
    {
      SSH_TRACE(2, ("Couldn't agree on kex or hostkey alg. (chosen_kex = "
                    "%s, chosen_host_key = %s)",
                    chosen_kex ? chosen_kex : "NULL",
                    chosen_host_key ? chosen_host_key : "NULL"));

      /* Failed to find acceptable kex method. */
      ssh_buffer_free(client_kexinit);
      ssh_buffer_free(server_kexinit);
      ssh_xfree(chosen_c_to_s_cipher);
      ssh_xfree(chosen_s_to_c_cipher);
      ssh_xfree(chosen_c_to_s_mac);
      ssh_xfree(chosen_s_to_c_mac);
      ssh_xfree(chosen_c_to_s_compression);
      ssh_xfree(chosen_s_to_c_compression);

      return FALSE;
    }

  ssh_buffer_free(client_kexinit);
  ssh_buffer_free(server_kexinit);

  if ((!chosen_c_to_s_cipher || !chosen_s_to_c_cipher ||
       !chosen_c_to_s_mac || !chosen_s_to_c_mac ||
       !chosen_c_to_s_compression || !chosen_s_to_c_compression))
    {
      /* Failed to agree on some algorithm. */

      ssh_xfree(chosen_c_to_s_cipher);
      ssh_xfree(chosen_s_to_c_cipher);
      ssh_xfree(chosen_c_to_s_mac);
      ssh_xfree(chosen_s_to_c_mac);
      ssh_xfree(chosen_c_to_s_compression);
      ssh_xfree(chosen_kex);
      ssh_xfree(chosen_host_key);
      return FALSE;
    }

  /* Set the selected algorithms. */
  ssh_tr_set_string(&tr->kex_name, chosen_kex);
  ssh_tr_set_string(&tr->host_key_name, chosen_host_key);
  ssh_tr_set_string(&tr->c_to_s.cipher_name, chosen_c_to_s_cipher);
  ssh_tr_set_string(&tr->s_to_c.cipher_name, chosen_s_to_c_cipher);
  ssh_tr_set_string(&tr->c_to_s.mac_name, chosen_c_to_s_mac);
  ssh_tr_set_string(&tr->s_to_c.mac_name, chosen_s_to_c_mac);
  ssh_tr_set_string(&tr->c_to_s.compression_name,
                    chosen_c_to_s_compression);
  ssh_tr_set_string(&tr->s_to_c.compression_name,
                    chosen_s_to_c_compression);

  SSH_TRACE(2, ("c_to_s: cipher %s, mac %s, compression %s",
                chosen_c_to_s_cipher, chosen_c_to_s_mac,
                chosen_c_to_s_compression));
  SSH_TRACE(2, ("s_to_c: cipher %s, mac %s, compression %s",
                chosen_s_to_c_cipher, chosen_s_to_c_mac,
                chosen_s_to_c_compression));

  /* 2.4.1 and older versions determine the correctness of the guess
     by comparing the chosen algorithms and their own preferred
     algorithms. */
  if (tr->ssh_old_guess_determination_bug_compat)
    {
      if (tr->server)
        {
          *guess_was_wrong = (strcmp(client_preferred_kex,
                                     chosen_kex) != 0 ||
                              strcmp(client_preferred_host_key,
                                     chosen_host_key) != 0);
        }
      else
        {
          *guess_was_wrong = (strcmp(server_preferred_kex,
                                     chosen_kex) != 0 ||
                              strcmp(server_preferred_host_key,
                                     chosen_host_key) != 0);
        }
    }

  ssh_xfree(client_preferred_kex);
  ssh_xfree(client_preferred_host_key);
  ssh_xfree(server_preferred_kex);
  ssh_xfree(server_preferred_host_key);

  SSH_DEBUG(4, ("Chosen host key algorithm: %s, Chosen kex algorithm: %s, %s",
                chosen_host_key, chosen_kex, *guess_was_wrong ?
                "Guessed wrong" : "Guessed right"));
  if (*guess_was_wrong)
    {
      SSH_DEBUG(4, ("Guessed host key algorithm: %s, "
                    "Guessed kex algorithm: %s",
                    tr->guessed_host_key, tr->guessed_kex));
    }

  /* Set the current kex method and hash handles (cipher, mac, and
     compression aren't changed until NEWKEYS is sent or received). */
  tr->kex = ssh_kex_lookup(chosen_kex);
  if (tr->kex == NULL)
    {
      ssh_fatal("ssh_tr_negotiate: chosen kex '%.100s' not found.",
                chosen_kex);
    }

  tr->hash = ssh_kex_allocate_hash(chosen_kex);
  if (tr->hash == NULL)
    {
      ssh_fatal("unable to allocate the hash function needed by the "
                "key exchange method %s.", chosen_kex);
    }

  /* Indicate that negotiation was successful. */
  return TRUE;
}

/* Gets the packet type from the buffer, without removing it.  The buffer
   should contain the packet payload. */

unsigned int ssh_tr_peek_packet_type(SshBuffer packet)
{
  return (unsigned int) (ssh_buffer_ptr(packet))[0];
}

/* Processes a received SSH_MSG_DISCONNECT packet.  Does not free the
   buffer. */

void ssh_tr_input_disconnect(SshTransportCommon tr, SshBuffer packet)
{
  unsigned int packet_type;
  SshUInt32 reason;
  char *message;

  SSH_DEBUG(4, ("Processing received SSH_MSG_DISCONNECT"));

  if (ssh_decode_buffer(packet,
                        SSH_FORMAT_CHAR, &packet_type,
                        SSH_FORMAT_UINT32, &reason,
                        SSH_FORMAT_UINT32_STR, &message, NULL,
                        SSH_FORMAT_UINT32_STR, NULL, NULL, /* language tag */
                        SSH_FORMAT_END) == 0)
    {
      SSH_TRACE(0, ("Bad SSH_MSG_DISCONNECT."));
      return;
    }
  SSH_ASSERT(packet_type == SSH_MSG_DISCONNECT);
  
  ssh_tr_up_disconnect(tr, FALSE, FALSE, reason, "%.300s", message);
  ssh_xfree(message);
}

/* Processes a received SSH_MSG_DEBUG packet.  Does not free the
   buffer. */

void ssh_tr_input_debug(SshTransportCommon tr, SshBuffer packet)
{
  unsigned int packet_type;
  char *message, *language;
  SshBufferStruct buffer;
  Boolean always_display;

  SSH_DEBUG(5, ("Processing received SSH_MSG_DEBUG."));

  if (ssh_decode_buffer(packet,
                        SSH_FORMAT_CHAR, &packet_type,
                        SSH_FORMAT_BOOLEAN, &always_display,
                        SSH_FORMAT_UINT32_STR, &message, NULL,
                        SSH_FORMAT_UINT32_STR, &language, NULL,
                        SSH_FORMAT_END) == 0)
    {
      SSH_TRACE(0, ("Bad SSH_MSG_DEBUG."));
      return;
    }
  SSH_ASSERT(packet_type == SSH_MSG_DEBUG);

  ssh_buffer_init(&buffer);
  ssh_encode_buffer(&buffer,
                    SSH_FORMAT_BOOLEAN, always_display,
                    SSH_FORMAT_UINT32_STR, message, strlen(message),
                    SSH_FORMAT_UINT32_STR, language, strlen(language),
                    SSH_FORMAT_END);
  ssh_tr_up_send(tr, SSH_CROSS_DEBUG, ssh_buffer_ptr(&buffer),
                 ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);
  ssh_xfree(message);
  ssh_xfree(language);
}

/* Tries to read packet data from the connection.  If a complete packet has
   been received, returns the packet payload in a buffer.  The caller is
   responsible for freeing the buffer with ssh_buffer_free().  Otherwise,
   this returns NULL.  If an error is received (such as EOF or a corrupted
   packet), this returns NULL and calls ssh_tr_up_disconnect() to
   pass the error to upper levels. */

SshBuffer ssh_tr_input_packet(SshTransportCommon tr)
{
  int len;
  unsigned int mac_len, pad_len, packet_type;
  unsigned char mac[SSH_MAX_HASH_DIGEST_LENGTH];
  SshBuffer packet;
  unsigned char *cp;
  unsigned char seq_buf[4];

  SSH_DEBUG(8, ("Reading packet from connection."));

 restart:

  packet = tr->incoming_packet;

  /* Allocate a packet buffer if appropriate. */
  if (packet == NULL)
    {
      packet = ssh_xbuffer_allocate(); /* This is a possible leak. XXX */
      tr->incoming_packet = packet;
      tr->incoming_packet_index = 0;
      tr->incoming_packet_len = 0;
      /* Reserve space for the initial 8 bytes. */
      ssh_xbuffer_append_space(packet, &cp, tr->incoming_granularity);
    }
  SSH_ASSERT(ssh_buffer_len(packet) >= tr->incoming_granularity);

  /* Read packet length if it has not yet been received. */
  while (tr->incoming_packet_index < tr->incoming_granularity)
    {
      len = ssh_stream_read(tr->connection,
                            ssh_buffer_ptr(packet) + tr->incoming_packet_index,
                            tr->incoming_granularity -
                            tr->incoming_packet_index);
      SSH_DEBUG(9, ("Read %d bytes", len));
      if (len < 0)
        return NULL;  /* No more data available at this time. */
      if (len == 0)
        {
          /* Received EOF. */
          SSH_TRACE(3, ("Received EOF."));
          if (tr->incoming_packet_index == 0)
            { /* Clean EOF at beginning of packet. */
              tr->up_outgoing_eof = TRUE;
              tr->received_state = RECEIVED_DEAD;
              ssh_tr_up_signal_input(tr);
            }
          else
            ssh_tr_up_disconnect(tr, TRUE, FALSE,
                                 SSH_DISCONNECT_CONNECTION_LOST,
                                 "Connection lost.");
          return NULL;
        }
      tr->incoming_packet_index += len;
    }

  /* Cache the length of incoming MAC. */
  mac_len = ssh_mac_length(tr->incoming_mac);

  /* Decrypt and retrieve packet length if it hasn't already been done. */

  if (tr->incoming_packet_len == 0)
    {
      /* Decrypt the first few bytes of the incoming packet. */
      if (ssh_cipher_transform(tr->incoming_cipher, ssh_buffer_ptr(packet),
                               ssh_buffer_ptr(packet),
                               tr->incoming_granularity) != SSH_CRYPTO_OK)
        /* XXX should disconnect. */
        ssh_fatal("ssh_tr_input_packet: decrypting length failed (gran %d)",
                  tr->incoming_granularity);

      /* Compute the total length of the packet. */

      tr->incoming_packet_len =
          SSH_GET_32BIT(ssh_buffer_ptr(packet)) + mac_len + 4;

      /* Sanity check the length. */
      if (tr->incoming_packet_len > SSH_MAX_TOTAL_PACKET_LENGTH)
        {
          /* Send a disconnect packet to the other side. */
          ssh_tr_up_disconnect(tr, TRUE, TRUE,
                               SSH_DISCONNECT_PROTOCOL_ERROR,
                               "Protocol error: packet too long: %d.",
                               tr->incoming_packet_len);
          tr->incoming_packet_len = 0; /* Just in case... */
          return NULL;
        }

      /* Reserve space for the entire packet. */
      ssh_xbuffer_append_space(packet, &cp,
                               tr->incoming_packet_len -
                               ssh_buffer_len(packet));
    }

  /* Keep reading until the entire packet has been received. */
  while (tr->incoming_packet_index < tr->incoming_packet_len)
    {
      len = ssh_stream_read(tr->connection,
                            ssh_buffer_ptr(packet) + tr->incoming_packet_index,
                            tr->incoming_packet_len -
                            tr->incoming_packet_index);
      SSH_DEBUG(9, ("Read %d bytes", len));
      if (len < 0)
        return NULL;  /* No more data available at this time. */
      if (len == 0)
        {
          /* Received EOF. */
          ssh_tr_up_disconnect(tr, TRUE, FALSE,
                               SSH_DISCONNECT_CONNECTION_LOST,
                               "Connection lost.");
          return NULL;
        }
      tr->incoming_packet_index += len;
    }

#ifdef DUMP_PACKETS
  SSH_DEBUG_HEXDUMP(3, ("Incoming packet:"),
                    ssh_buffer_ptr(packet),
                    ssh_buffer_len(packet));
#endif /* DUMP_PACKETS */

  /* All of the packet has now been received.  */
  if ((tr->incoming_packet_len - mac_len) % tr->incoming_granularity != 0)
    {
      ssh_tr_up_disconnect(tr, TRUE, FALSE, SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Received packet with wrong granularity.");
      return NULL;
    }

  /* Decrypt the rest of the packet (the first cipher block has already been
     decrypted). */

  if (ssh_cipher_transform(tr->incoming_cipher,
                           ssh_buffer_ptr(packet) + tr->incoming_granularity,
                           ssh_buffer_ptr(packet) + tr->incoming_granularity,
                           tr->incoming_packet_len -
                           tr->incoming_granularity - mac_len)
      != SSH_CRYPTO_OK)
    /* Should disconnect. */
    ssh_fatal("ssh_tr_input_packet: decrypting rest failed "
              "(len %d gran %d mac %d)",
          tr->incoming_packet_len, tr->incoming_granularity, mac_len);

  /* Verify MAC. */

  if (!tr->ssh_old_mac_bug_compat)
    {
      /* Everything is a-ok */
      ssh_mac_start(tr->incoming_mac);
      SSH_PUT_32BIT(seq_buf, tr->incoming_sequence_number);
      ssh_mac_update(tr->incoming_mac, seq_buf, 4);
      ssh_mac_update(tr->incoming_mac, ssh_buffer_ptr(packet),
                     tr->incoming_packet_len - mac_len);
      ssh_mac_final(tr->incoming_mac, mac);
    }
  else /* other side counts the MAC in the old (==wrong) way. */
    {
      ssh_mac_start(tr->incoming_mac);
      ssh_mac_update(tr->incoming_mac, ssh_buffer_ptr(packet),
                     tr->incoming_packet_len - mac_len);
      SSH_PUT_32BIT(seq_buf, tr->incoming_sequence_number);
      ssh_mac_update(tr->incoming_mac, seq_buf, 4);
      ssh_mac_final(tr->incoming_mac, mac);
    }

  if (memcmp(mac, ssh_buffer_ptr(packet) + tr->incoming_packet_len - mac_len,
             mac_len) != 0)
    {
      /* MAC fails. */
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_MAC_ERROR,
                           "Message authentication check fails.");
      return NULL;
    }

  /* MAC ok.  Remove the MAC from the packet. */
  ssh_buffer_consume_end(packet, mac_len);

  if (ssh_decode_buffer(packet,
                        SSH_FORMAT_UINT32, NULL,
                        SSH_FORMAT_CHAR, &pad_len,
                        SSH_FORMAT_END) == 0)
    {
      ssh_tr_up_disconnect(tr, TRUE, TRUE, SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Badly formatted packet");
      return NULL;
    }
  if (pad_len > ssh_buffer_len(packet))
    {
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Bad padding length %d", pad_len);
      return NULL;
    }
  ssh_buffer_consume_end(packet, pad_len);
  /* At this point, the buffer contains the (possibly compressed) payload. */

  /* Uncompress the payload if appropriate. */
  tr->compressed_incoming_bytes += ssh_buffer_len(packet);
  if (!ssh_compress_is_none(tr->compression_incoming))
    {
      SshBuffer aux_packet;

      /* Uncompress the data. */
      ssh_buffer_clear(tr->compression_buffer);
      ssh_compress_buffer(tr->compression_incoming, ssh_buffer_ptr(packet),
                          ssh_buffer_len(packet), tr->compression_buffer);

      /* Swap the buffers so that uncompressed data is returned.  We'll reuse
         the other buffer for compression later. */
      aux_packet = tr->compression_buffer;
      tr->compression_buffer = packet;
      packet = aux_packet;
    }
  tr->uncompressed_incoming_bytes += ssh_buffer_len(packet);

  /* Update packet sequence number. */
  tr->incoming_sequence_number++;

  /* Return the payload.  The caller will free the packet. */
  tr->incoming_packet = NULL;

#ifdef DUMP_PACKETS
  SSH_DEBUG_HEXDUMP(3, ("Incoming packet -- decrypted --:"),
                    ssh_buffer_ptr(packet),
                    ssh_buffer_len(packet));
#endif /* DUMP_PACKETS */

  /* Check for SSH_MSG_IGNORE packets. */
  packet_type = ssh_tr_peek_packet_type(packet);
  switch (packet_type)
    {
    case SSH_MSG_IGNORE:      /* These packets are immediately ignored. */
      ssh_buffer_free(packet);
      goto restart;

    case SSH_MSG_DISCONNECT:  /* The other side is disconnecting. */
      ssh_tr_input_disconnect(tr, packet);
      ssh_buffer_free(packet);
      return NULL;

    case SSH_MSG_DEBUG:       /* The other side sends a debug packet. */
      ssh_tr_input_debug(tr, packet);
      ssh_buffer_free(packet);
      goto restart;

    case SSH_MSG_UNIMPLEMENTED: /* Strange, our packet was unimplemented? */
      /* XXX If, for example, we receive this when waiting for a
         response to SERVICE_REQUEST packet, or after sending
         SERVICE_ACCEPT message, we will hang to select. Which is not
         nice. */
      SSH_TRACE(0, ("Strange, other side indicates our message as "
                    "unimplemented."));
      ssh_buffer_free(packet);
      return NULL;
    default:
      /* Other packet types are processed normally. */
      break;
    }

  /* Return the packet for normal processing. */
  return packet;
}

Boolean ssh_tr_process_received_kexinit(SshTransportCommon tr,
                                        SshBuffer packet)
{
  Boolean guess_was_wrong = TRUE;

  SSH_DEBUG(4, ("Processing received SSH_MSG_KEXINIT."));

  /* Save the kexinit packet. */
  if (tr->server)
    {
      SSH_ASSERT(tr->client_kexinit_packet == NULL);
      tr->client_kexinit_packet = packet;
    }
  else
    {
      SSH_ASSERT(tr->server_kexinit_packet == NULL);
      tr->server_kexinit_packet = packet;
    }

  /* Perform algorithm negotiation computations based on the two kex
     packets. */
  if (!ssh_tr_negotiate(tr, &guess_was_wrong))
    {
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_KEY_EXCHANGE_FAILED,
                           "Algorithm negotiation failed.");
      return FALSE;
    }

  if (guess_was_wrong && tr->first_kex_packet_follows)
    {
      tr->received_state = RECEIVED_KEXINIT;
    }
  else
    {
      /* If we aren't expecting to receive kex1, go directly to expecting
         kex2. */
      if ((tr->server ? tr->kex->server_input_kex1 :
           tr->kex->client_input_kex1)
          == NULL_FNPTR)
        tr->received_state = RECEIVED_KEX1_FINAL;
      else
        tr->received_state = RECEIVED_KEX1_IGNORED;
    }

  SSH_ASSERT(tr->sent_state == SENT_KEXINIT);
  if (guess_was_wrong || (!tr->server && !tr->send_kex1_immediately))
    {
      /* Resend our KEX1 packet, if we are to send one. */
      packet = ssh_tr_make_kex1(tr);
      if (packet)
        {
          ssh_tr_send_packet(tr, ssh_buffer_ptr(packet),
                             ssh_buffer_len(packet));
          /* Note that the packet has been saved for key negotiation, and
             cannot be freed here. */
        }
    }

  tr->sent_state = SENT_KEX1_FINAL;

  return TRUE;
}

/* Receives a kexinit packet from the remote host.  Returns TRUE if
   reading should continue. */
Boolean ssh_tr_input_kexinit(SshTransportCommon tr)
{
  int packet_type;
  SshBuffer packet;

  SSH_DEBUG(5, ("Getting a SSH_MSG_KEXINIT packet from connection."));

  /* If we haven't sent our kexinit yet, do nothing (the send will
     probably just involve buffering, and should happen fast).  This simplifies
     code below, since we can now always perform negotiation in this
     function. */
  if (tr->sent_state < SENT_KEXINIT)
    return FALSE;

  /* Read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return FALSE;

  /* If we have requested rekey, we might get here with any normal data
     packet until the other side responds with a KEXINIT packet.  If the packet
     is a data packet, and doing rekey, we process it as a normal packet. */

  /* Check that the packet type is KEXINIT. */
  packet_type = ssh_tr_peek_packet_type(packet);
  if (packet_type != SSH_MSG_KEXINIT)
    {
      ssh_buffer_free(packet);
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Received packet type %d, expecting KEXINIT.",
                           (int)packet_type);
      return FALSE;
    }

  /* Process the received packet now.  This code is shared with rekey. */
  return ssh_tr_process_received_kexinit(tr, packet);
}

/* Send a packet containing only the packet type.  This does not update
   state. */
void ssh_tr_send_simple_packet(SshTransportCommon tr,
                               unsigned int packet_type)
{
  SshBufferStruct buffer;

  SSH_DEBUG(9, ("Sending packet. Type %d.", packet_type));

  ssh_buffer_init(&buffer);
  ssh_bufaux_put_char(&buffer, packet_type);
  ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer),
                     ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);
}

/* Takes the new encryption keys into use for one direction.  This clears
   the keys from memory after they have been set. */

void ssh_tr_set_keys(SshTransportCommon tr, struct SideKexInfo *info,
                     size_t *granularityp, Boolean is_outgoing,
                     SshCipher *cipherp, SshMac *macp,
                     SshCompression *compressionp)
{
  size_t key_len, mac_len;

  SSH_DEBUG(4, ("Setting new keys and algorithms"));

  /* Set encryption algorithm. */

  if (*cipherp)
    ssh_cipher_free(*cipherp);

  key_len = ssh_cipher_get_key_length(info->cipher_name);

  if (key_len == 0)
    {
      if ((strcasecmp("twofish", info->cipher_name) == 0) ||
          (strncasecmp("twofish-", info->cipher_name, 8) == 0))
        key_len = 32;  /* Twofish uses 256 bit key. */
      else
        key_len = 16;  /* Default is 128 bits if variable length keys. */
    }
  SSH_ASSERT(key_len <= sizeof(info->encryption_key));

  SSH_DEBUG(3, ("Allocating cipher: name: %s, key_len: %d.", info->cipher_name,
                key_len));
  if (ssh_cipher_allocate(info->cipher_name, info->encryption_key, key_len,
                          is_outgoing, cipherp) != SSH_CRYPTO_OK)
    ssh_fatal("ssh_tr_set_keys: cipher init failed: %.100s",
              info->cipher_name);
  *granularityp = ssh_cipher_get_block_length(*cipherp);
  if (*granularityp < 8)
    *granularityp = 8;

  /* Set iv.  We intentionally ignore the return status in case stream
     ciphers return an error. */
  ssh_cipher_set_iv(*cipherp, info->iv);

  /* Set mac algorithm. */
  if (*macp)
    ssh_mac_free(*macp);

  if (!tr->ssh_old_sha1_key_len_bug_compat)
    {
      if (!strncmp(info->mac_name, "hmac-sha1", strlen("hmac-sha1")) ||
          !strncmp(info->mac_name, "hmac-sha", strlen("hmac-sha")))
        mac_len = 20;
      else
        mac_len = 16;
    }
  else
    {
      SSH_DEBUG(3, ("MAC length forced to 16 bytes, due to incompatibility "
                    "in older version."));
      mac_len = 16;
    }

  if (ssh_mac_allocate(info->mac_name, info->integrity_key, mac_len,
                       macp) != SSH_CRYPTO_OK)
    ssh_fatal("ssh_tr_set_keys: mac init failed: %.100s", info->mac_name);

  /* Set compression algorithm.  First we free any old compression state. */
  if (*compressionp)
    ssh_compress_free(*compressionp);
  *compressionp = ssh_compress_allocate(info->compression_name,
                                        tr->params->compression_level,
                                        is_outgoing);
  if (!*compressionp)
    ssh_fatal("ssh_tr_set_keys: compression init failed: %.100s",
              info->compression_name);

  /* Clear the keys from memory. */
  memset(info->encryption_key, 0, sizeof(info->encryption_key));
  memset(info->iv, 0, sizeof(info->iv));
  memset(info->integrity_key, 0, sizeof(info->integrity_key));
}

/* Called when the validity check for the received host key is complete.
   If the validity check failed, this disconnects; otherwise this proceeds
   with the key exchange. */
void ssh_tr_key_check_done(Boolean result, void *context)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(5, ("Key check finalized. Key is %saccepted.",
                result ? "" : "not "));

  SSH_ASSERT(tr->received_state == RECEIVED_KEY_CHECK);

  if (!result)
    {
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_HOST_KEY_NOT_VERIFIABLE,
                           "Could not ascertain validity of host key.");
      return;
    }

  /* The host key has been accepted. */

  /* Start creating the KEX2 packet and send it in the callback. */
  ssh_tr_make_kex2(tr, ssh_tr_key_check_done_make_kex2_cb);
  return;
}

/* This function gets the constructed KEX2 packet from the lower level
   and sends it. */

void ssh_tr_key_check_done_make_kex2_cb(SshBuffer packet,
                                        SshTransportCommon tr)
{  
  if (packet)
    {
      ssh_tr_send_packet(tr, ssh_buffer_ptr(packet),
                         ssh_buffer_len(packet));
      /* Note that KEX2 packets are not saved anywhere and need to be freed
         here. */
      ssh_buffer_free(packet);
    }

  /* Send a NEWKEYS packet. */
  ssh_tr_send_simple_packet(tr, SSH_MSG_NEWKEYS);
  tr->sent_state = SENT_NEWKEYS;

  /* Take the new keys and algorithms into use. */
  if (tr->server)
    ssh_tr_set_keys(tr, &tr->s_to_c, &tr->outgoing_granularity, TRUE,
                    &tr->outgoing_cipher, &tr->outgoing_mac,
                    &tr->compression_outgoing);
  else
    ssh_tr_set_keys(tr, &tr->c_to_s, &tr->outgoing_granularity, TRUE,
                    &tr->outgoing_cipher, &tr->outgoing_mac,
                    &tr->compression_outgoing);

  /* Go directly to expecting newkeys if not expecting to receive KEX2. */
  if ((tr->server ? tr->kex->server_input_kex2 : tr->kex->client_input_kex2)
      == NULL_FNPTR)
    tr->received_state = RECEIVED_KEX2;
  else
    tr->received_state = RECEIVED_KEX1_FINAL;

  ssh_tr_process_output(tr);
  /* Make sure the input callback is called, because the signing might
     have been asynchoronous, and we haven't called
     ssh_tr_process_input() yet. */
  tr->read_has_blocked = TRUE;
  ssh_tr_wake_up_input(tr);
}

/* Receive a kex1 packet.  If appropriate, send kex2 in response.
   Returns TRUE if processing should continue, and FALSE if there is
   nothing more to do. */
Boolean ssh_tr_input_kex1(SshTransportCommon tr)
{
  SshBuffer packet;
  Boolean result;

  SSH_DEBUG(5, ("Receiving first key exchange packet."));

  /* Don't proceed until we have sent our own final KEX1 packet.  This
     simplifies the code. */
  if (tr->sent_state < SENT_KEX1_FINAL)
    return FALSE;

  /* Read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return FALSE;

  /* Ignore the kex1 packet if appropriate. */
  if (tr->received_state == RECEIVED_KEXINIT)
    {
      /* We are receiving a kex1 packet that needs to be ignored. */
      ssh_buffer_free(packet);
      /* If we aren't expecting to receive kex1, go directly to expecting
         kex2. */
      if ((tr->server ? tr->kex->server_input_kex1 :
           tr->kex->client_input_kex1)
          == NULL_FNPTR)
        tr->received_state = RECEIVED_KEX1_FINAL;
      else
        tr->received_state = RECEIVED_KEX1_IGNORED;
      return TRUE;
    }

  /* Save the received KEX1 packet. */
  if (tr->server)
    result = (*tr->kex->server_input_kex1)(tr, packet);
  else
    result = (*tr->kex->client_input_kex1)(tr, packet);

  if (!result)
    {
      ssh_buffer_free(packet);
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Malformed first key exhange packet.");
      return FALSE;  /* Parsing failed. */
    }

  /* Mark that we are processing key check. */
  tr->received_state = RECEIVED_KEY_CHECK;

  if (tr->key_check)
    (*tr->key_check)(tr->server_host_name,
                     ssh_buffer_ptr(tr->public_host_key_blob),
                     ssh_buffer_len(tr->public_host_key_blob),
                     (unsigned char*) tr->host_key_name,
                     ssh_tr_key_check_done,
                     (void *)tr,
                     tr->key_check_context);
  else
    ssh_tr_key_check_done(TRUE, (void *)tr);

  ssh_buffer_free(packet);
  return TRUE;
}

/* Forward declaration. */
void ssh_tr_input_kex2_finalize(SshTransportCommon tr);

/* Process KEX2 packet.  If no packet is yet available, return. Rest
   is done ssh_tr_input_kex2_finalize(), after key check.  */
void ssh_tr_input_kex2(SshTransportCommon tr)
{
  SshBuffer packet;

  SSH_DEBUG(5, ("Receiving second key exchange packet."));

  /* Read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return;

  /* Process the packet to finalize our part of the key exchange.
     This will also free the packet. */
  if (tr->server)
    (*tr->kex->server_input_kex2)(tr, packet,
                                  ssh_tr_input_kex2_finalize);
  else
    (*tr->kex->client_input_kex2)(tr, packet,
                                  ssh_tr_input_kex2_finalize);

}

/* If key exchange fails, initiate disconnect and return.  If KEX2 is
   successfully received and the key exchange successfully completes,
   update state and send NEWKEYS. After these, this calls
   ssh_tr_process_input() to continue the protocol. */
void ssh_tr_input_kex2_finalize(SshTransportCommon tr)
{
  Boolean success;

  success = tr->key_check_result;

  if (!success)
    {
      /* Disconnect. */
      ssh_tr_up_disconnect(tr, TRUE, TRUE, SSH_DISCONNECT_KEY_EXCHANGE_FAILED,
                           "Key exchange failed.");
      return;
    }

  /* Key exchange has been successful.  Send newkeys. */
  ssh_tr_send_simple_packet(tr, SSH_MSG_NEWKEYS);

  /* Take the new keys and algorithms into use. */
  if (tr->server)
    ssh_tr_set_keys(tr, &tr->s_to_c, &tr->outgoing_granularity, TRUE,
                    &tr->outgoing_cipher, &tr->outgoing_mac,
                    &tr->compression_outgoing);
  else
    ssh_tr_set_keys(tr, &tr->c_to_s, &tr->outgoing_granularity, TRUE,
                    &tr->outgoing_cipher, &tr->outgoing_mac,
                    &tr->compression_outgoing);

  /* Update state. */
  tr->sent_state = SENT_NEWKEYS;
  tr->received_state = RECEIVED_KEX2;
  ssh_tr_process_output(tr);

  ssh_tr_process_input(tr);
}

/* Process NEWKEYS packet.  This will cause new algorithms to be taken
   into use for incoming packets. */

Boolean ssh_tr_input_newkeys(SshTransportCommon tr)
{
  SshBuffer packet;
  unsigned int packet_type;

  SSH_DEBUG(4, ("Receiving SSH_MSG_NEWKEYS."));

  /* Read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return FALSE;
  packet_type = ssh_tr_peek_packet_type(packet);
  if (packet_type != SSH_MSG_NEWKEYS)
    {
      ssh_tr_up_disconnect(tr, TRUE, TRUE, SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Protocol error: Received %d as newkeys",
                           packet_type);
      return FALSE;
    }
  ssh_buffer_free(packet);

  if (tr->server)
    {
      ssh_tr_set_keys(tr, &tr->c_to_s, &tr->incoming_granularity, FALSE,
                      &tr->incoming_cipher, &tr->incoming_mac,
                      &tr->compression_incoming);
      
#ifdef SSHDIST_ZLIB_LEVEL_DETECTION
      /* Sync the two directions of transport layer here.  If client
         has requested compression and selects a compression level,
         the server will use roughly the same compression level after
         a short while. */
      
      ssh_compress_sync_levels(tr->compression_incoming,
                               tr->compression_outgoing);
#endif /* SSHDIST_ZLIB_LEVEL_DETECTION */
    }
  else
    ssh_tr_set_keys(tr, &tr->s_to_c, &tr->incoming_granularity, FALSE,
                    &tr->incoming_cipher, &tr->incoming_mac,
                    &tr->compression_incoming);

  /* Mark that we have received NEWKEYS. */
  tr->received_state = RECEIVED_NEWKEYS;

  return TRUE;
}

/* Send a SSH_CROSS_STARTUP packet upstream. */

void ssh_tr_up_send_startup(SshTransportCommon tr)
{
  SshBufferStruct buffer;
  char buf[100];

  SSH_DEBUG(4, ("Sending startup packet to application layer."));

  ssh_buffer_init(&buffer);
  ssh_bufaux_put_uint32_string(&buffer, SSH_CROSS_LAYER_VERSION,
                               strlen(SSH_CROSS_LAYER_VERSION));
  ssh_bufaux_put_uint32_string(&buffer, tr->session_identifier,
                               tr->session_identifier_len);
  ssh_bufaux_put_uint32_string(&buffer, tr->remote_version,
                               strlen(tr->remote_version));
  if (ssh_tcp_get_remote_address(tr->connection, buf, sizeof(buf)))
    ssh_bufaux_put_uint32_string(&buffer, buf, strlen(buf));
  else
    ssh_bufaux_put_uint32_string(&buffer, NULL, 0);
  if (ssh_tcp_get_remote_port(tr->connection, buf, sizeof(buf)))
    ssh_bufaux_put_uint32_string(&buffer, buf, strlen(buf));
  else
    ssh_bufaux_put_uint32_string(&buffer, NULL, 0);
  ssh_tr_up_send(tr, SSH_CROSS_STARTUP,
                 ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);

  /* Mix the session identifier into our random number generator.  The
     session identifier depends on the entire key exchange (every bit
     of it).  It thus combines entropy from both the local machine and
     the remote host.  Some of the entropy is secret to an outsider
     (e.g., the secret session key).  The random state can be used to
     update the random seed file later. */
  ssh_random_add_noise(tr->session_identifier,
                       tr->session_identifier_len);
  ssh_random_stir();
}

/* Send a SSH_CROSS_ALGORITHMS packet upstream. */

void ssh_tr_up_send_algorithms(SshTransportCommon tr)
{
  SshBufferStruct buffer;
  SshBuffer pub_host_key_buf;

  SSH_DEBUG(4, ("Sending algorithms to application layer."));

  if (tr->server)
    {
      pub_host_key_buf =
        ssh_host_key_get_pubkey_by_algname(tr->host_key_name,
                                           tr->host_keys_ctx);
    }
  else
    {
      pub_host_key_buf = tr->public_host_key_blob;
    }

  SSH_ASSERT(pub_host_key_buf != NULL);

  ssh_buffer_init(&buffer);
  ssh_bufaux_put_uint32_string(&buffer, tr->kex_name, strlen(tr->kex_name));
  ssh_bufaux_put_uint32_string(&buffer, tr->host_key_name,
                               strlen(tr->host_key_name));
  ssh_bufaux_put_uint32_string(&buffer, ssh_buffer_ptr(pub_host_key_buf),
                               ssh_buffer_len(pub_host_key_buf));
  ssh_bufaux_put_uint32_string(&buffer, tr->c_to_s.cipher_name,
                               strlen(tr->c_to_s.cipher_name));
  ssh_bufaux_put_uint32_string(&buffer, tr->s_to_c.cipher_name,
                               strlen(tr->s_to_c.cipher_name));
  ssh_bufaux_put_uint32_string(&buffer, tr->c_to_s.mac_name,
                               strlen(tr->c_to_s.mac_name));
  ssh_bufaux_put_uint32_string(&buffer, tr->s_to_c.mac_name,
                               strlen(tr->s_to_c.mac_name));
  ssh_bufaux_put_uint32_string(&buffer, tr->c_to_s.compression_name,
                               strlen(tr->c_to_s.compression_name));
  ssh_bufaux_put_uint32_string(&buffer, tr->s_to_c.compression_name,
                               strlen(tr->s_to_c.compression_name));
  ssh_tr_up_send(tr, SSH_CROSS_ALGORITHMS,
                 ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);
}

/* Try to receive a service request packet.  If received, call the
   application callback to determine whether the service can be accepted.
   If the service is denied, send disconnect and destroy this context.
   If the service is accepted, send a service accept packet, and enter
   interactive mode. */

Boolean ssh_tr_input_service_request(SshTransportCommon tr)
{
  SshBuffer packet;
  SshBufferStruct buffer;
  unsigned int packet_type;

  SSH_DEBUG(5, ("Waiting for a service request packet."));

  /* Try to read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return FALSE;

  if (ssh_decode_buffer(packet,
                        SSH_FORMAT_CHAR, &packet_type,
                        SSH_FORMAT_UINT32_STR, &tr->service_name, NULL,
                        SSH_FORMAT_END) == 0 ||
      packet_type != SSH_MSG_SERVICE_REQUEST)
    {
      /* We shouldn't have received this here.  This is a protocol error. */
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Protocol error: bad service request %d",
                           packet_type);
      return FALSE;
    }

  /* Free the packet. */
  ssh_buffer_free(packet);

  /* Send the service request upstream. */
  ssh_buffer_init(&buffer);
  ssh_bufaux_put_uint32_string(&buffer, tr->service_name,
                               strlen(tr->service_name));
  ssh_tr_up_send(tr, SSH_CROSS_SERVICE_REQUEST,
                 ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);

  tr->received_state = RECEIVED_SERVICE_REQUEST;

  return TRUE;
}

/* Try to receive a service accept packet.  Advance to interactive state
   when received. */

Boolean ssh_tr_input_service_accept(SshTransportCommon tr)
{
  SshBuffer packet;
  unsigned int packet_type;
  char *service_name = NULL;
  size_t service_name_length;

  SSH_PRECOND(!tr->server);

  SSH_DEBUG(4, ("Waiting for a service accept packet."));

  /* Try to read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return FALSE;

  /* Check that it is a service accept packet. */
  if (!tr->ssh_old_msg_service_accept_bug_compat)
    {
      if (ssh_decode_buffer(packet,
                            SSH_FORMAT_CHAR, &packet_type,
                            SSH_FORMAT_UINT32_STR, &service_name,
                            &service_name_length,
                            SSH_FORMAT_END) == 0 ||
          packet_type != SSH_MSG_SERVICE_ACCEPT)
        {
          /* We shouldn't have received this here.  This is a protocol
             error. */
          ssh_tr_up_disconnect(tr, TRUE, TRUE,
                               SSH_DISCONNECT_PROTOCOL_ERROR,
                               "Protocol error: bad service accept %d",
                               packet_type);
          return FALSE;
        }

      SSH_DEBUG(4, ("Received SSH_MSG_SERVICE_ACCEPT with service "
                    "name \"%s\".", service_name));

      ssh_xfree(service_name);
    }
  else
    {
      if (ssh_decode_buffer(packet,
                            SSH_FORMAT_CHAR, &packet_type,
                            SSH_FORMAT_END) == 0 ||
          packet_type != SSH_MSG_SERVICE_ACCEPT)
        {
          /* We shouldn't have received this here.  This is a protocol
             error. */
          ssh_tr_up_disconnect(tr, TRUE, TRUE,
                               SSH_DISCONNECT_PROTOCOL_ERROR,
                               "Protocol error: bad service accept %d",
                               packet_type);
          return FALSE;
        }
      SSH_DEBUG(4, ("Received SSH_MSG_SERVICE_ACCEPT."));
    }

  /* Free the packet. */
  ssh_buffer_free(packet);

  /* Send a SSH_CROSS_STARTUP message upwards. */
  ssh_tr_up_send_startup(tr);

  /* Send a SSH_CROSS_ALGORITHMS packet upwards. */
  ssh_tr_up_send_algorithms(tr);

  /* Update state. */
  tr->received_state = RECEIVED_INTERACTIVE;

  /* Process output, so that we get advanced from the service request state
     if we are still lingering there. */
  ssh_tr_process_output(tr);

  return TRUE;
}

char *tr_mangle_cipher_names(SshTransportCommon tr, char *ciphers)
{
  const char *p;
  char *new_list = NULL, *t, *to_be_deleted;

  if (!tr->ssh_old_des_cbc_naming_bug_compat)
    return ciphers; /* No mangling was required. */
  
  for (p = ciphers; p != NULL; p = ssh_snlist_step_forward(p))
    {
      t = ssh_snlist_get_name(p);
      SSH_ASSERT(t != NULL);
      to_be_deleted = new_list;
      if (strcmp(t, "des-cbc@ssh.com") == 0)
        {          
          ssh_xfree(t);
          t = ssh_xstrdup("des-cbc");
        }

      if (new_list == NULL)
        new_list = t;
      else
        ssh_xdsprintf(&new_list, "%s,%s", new_list, t);

      ssh_xfree(to_be_deleted);
    }

  SSH_POSTCOND(new_list != NULL);
  return new_list;
}

/* Constructs and sends an outgoing KEXINIT packet with the explicitly
   given algorithms.  This does not update state. */

void ssh_tr_output_kexinit_explicit(SshTransportCommon tr,
                                    const char *ciphers_c_to_s,
                                    const char *ciphers_s_to_c,
                                    const char *macs_c_to_s,
                                    const char *macs_s_to_c,
                                    const char *compressions_c_to_s,
                                    const char *compressions_s_to_c,
                                    const char *host_key_algorithms)
{
  int i;
  SshBuffer packet, kex1_packet = NULL;
  char *cp, *cp2;

  /* Construct our kex1 packet so that we know whether we are supposed to
     send it as a guessed packet for our default method. */
  /* If the other end doesn't handle guesses properly, don't send the
     kex1 packet at this stage. */
  if (tr->send_kex1_immediately)
    kex1_packet = ssh_tr_make_kex1(tr);

  /* Construct the outgoing KEXINIT packet. */
  packet = ssh_xbuffer_allocate();

  ssh_bufaux_put_char(packet, SSH_MSG_KEXINIT);
  for (i = 0; i < 16; i++)
    ssh_bufaux_put_char(packet, ssh_random_get_byte());

  cp2 = ssh_kex_get_supported();
  cp = ssh_snlist_intersection(tr->params->kex_algorithms, cp2);
  ssh_xfree(tr->guessed_kex);
  tr->guessed_kex = ssh_snlist_get_name(cp);
  ssh_xfree(cp2);
  SSH_ASSERT(cp != NULL);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: kex algs = %s", cp));
  ssh_xfree(cp);

  cp = ssh_host_key_get_supported(FALSE);
  cp2 = ssh_snlist_intersection(host_key_algorithms, cp);
  ssh_xfree(tr->guessed_host_key);
  tr->guessed_host_key = ssh_snlist_get_name(cp2);
  ssh_xfree(cp);
  SSH_ASSERT(cp2 != NULL);
  ssh_bufaux_put_uint32_string(packet, cp2, strlen(cp2));
  SSH_TRACE(5, ("local kexinit: host key algs = %s", cp2));
  ssh_xfree(cp2);

  cp = ssh_snlist_intersection_cipher(ciphers_c_to_s);
  SSH_ASSERT(cp != NULL);
  /* mangle des-cbc@ssh.com to des-cbc, if needed. */
  cp = tr_mangle_cipher_names(tr, cp);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: ciphers c to s = %s", cp));
  ssh_xfree(cp);

  cp = ssh_snlist_intersection_cipher(ciphers_s_to_c);
  SSH_ASSERT(cp != NULL);
  /* mangle des-cbc@ssh.com to des-cbc, if needed. */
  cp = tr_mangle_cipher_names(tr, cp);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: ciphers s to c = %s", cp));
  ssh_xfree(cp);

  cp = ssh_snlist_intersection_mac(macs_c_to_s);
  SSH_ASSERT(cp != NULL);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: macs c to s = %s", cp));
  ssh_xfree(cp);

  cp = ssh_snlist_intersection_mac(macs_s_to_c);
  SSH_ASSERT(cp != NULL);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: macs s to c = %s", cp));
  ssh_xfree(cp);

  cp = ssh_snlist_intersection_compression(compressions_c_to_s);
  SSH_ASSERT(cp != NULL);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: compressions c to s = %s", cp));
  ssh_xfree(cp);

  cp = ssh_snlist_intersection_compression(compressions_s_to_c);
  SSH_ASSERT(cp != NULL);
  ssh_bufaux_put_uint32_string(packet, cp, strlen(cp));
  SSH_TRACE(5, ("local kexinit: compressions s to c = %s", cp));
  ssh_xfree(cp);

  /* Put language-strings to packet as empty strings.
     This is legal according to specs. */

  ssh_bufaux_put_uint32_string(packet, "", 0);
  ssh_bufaux_put_uint32_string(packet, "", 0);

  ssh_bufaux_put_boolean(packet, kex1_packet != NULL);
  SSH_TRACE(4, ("local kexinit: first_packet_follows = %s",
                kex1_packet != NULL ? "TRUE" : "FALSE"));

  for (i = 0; i < 4; i++)
    ssh_bufaux_put_char(packet, 0);

  /* Save the packet for key id calculation. */
  if (tr->server)
    {
      SSH_ASSERT(tr->server_kexinit_packet == NULL);
      tr->server_kexinit_packet = packet;
    }
  else
    {
      SSH_ASSERT(tr->client_kexinit_packet == NULL);
      tr->client_kexinit_packet = packet;
    }

  /* Send the packet to the outgoing stream. */
  ssh_tr_send_packet(tr, ssh_buffer_ptr(packet), ssh_buffer_len(packet));

  /* If we are to send a guessed KEX1 packet, send it now.  Note that the
     packet has already been saved for key negotiation. */
  if (kex1_packet != NULL)
    ssh_tr_send_packet(tr, ssh_buffer_ptr(kex1_packet),
                       ssh_buffer_len(kex1_packet));
}

/* Constructs and sends an outgoing KEXINIT packet with the algorithms in
   the params structure.  This does not update state. */

void ssh_tr_output_kexinit(SshTransportCommon tr)
{
  ssh_tr_output_kexinit_explicit(tr,
                                 tr->params->ciphers_c_to_s,
                                 tr->params->ciphers_s_to_c,
                                 tr->params->macs_c_to_s,
                                 tr->params->macs_s_to_c,
                                 tr->params->compressions_c_to_s,
                                 tr->params->compressions_s_to_c,
                                 tr->params->host_key_algorithms);
}

/* Initiate rekey after KEXINIT was received in interactive mode.  This will
   send our own kexinit packet in response.  The kexinit packet received
   as argument is saved for further processing. */

void ssh_tr_input_start_rekey(SshTransportCommon tr, SshBuffer packet)
{
  SSH_DEBUG(5, ("Starting rekey."));

  /* We can get here in either of two states:
     1. We have requested to start rekey, and the other side is now
     replying.  (Note: it is possible that the requests cross; that case
     is indistinguishable from this case.)
     2. The other side is initiating rekey.
     The first case is identified by rekey already being true.
     Initialize for rekey unless we have already started it on our part. */
  if (!tr->rekey_request_sent)
    {
      /* Initialize state for rekey. */

      /* Notify application level of the started rekey. */
      if (tr->rekey_notify_callback)
        (*tr->rekey_notify_callback)(tr->rekey_notify_context);

      /* Cleanup the remains of any previous key exchange. */
      ssh_tr_kex_cleanup(tr);

      /* Mark that we are doing a rekey. */
      tr->doing_rekey = TRUE;

      SSH_ASSERT(tr->sent_state == SENT_INTERACTIVE);
      SSH_ASSERT(tr->received_state == RECEIVED_INTERACTIVE);

      tr->sent_state = SENT_VERSION;

      /* Send our own kexinit packet (possibly with guessed kex1).  This does
         not update state. */
      ssh_tr_output_kexinit(tr);

      /* Update send state. */
      tr->sent_state = SENT_KEXINIT;
    }

  /* Process the received kexinit packet identically to what we did during
     the initial key exchange.  This will update received_state. */
  ssh_tr_process_received_kexinit(tr, packet);
}

/* Try to receive and process a packet in interactive mode.  This returns
   FALSE if no packet is available or an error occurs, and TRUE if processing
   should continue.  If we return FALSE without trying to read, we will set
   the read_has_blocked flag to indicate that reads must be explicitly
   restarted. */

Boolean ssh_tr_input_interactive(SshTransportCommon tr)
{
  SshBuffer packet;
  unsigned int packet_type;
  SshBufferStruct buffer;
  unsigned char seq_buf[4];

  /* If the queue of packets going up is too long, don't read any more packets
     until it has drained. */
  if (ssh_buffer_len(&tr->up_outgoing) >
      XMALLOC_MAX_SIZE - SSH_CONTROL_RESERVE - SSH_MAX_PAYLOAD_LENGTH ||
      ssh_buffer_len(&tr->up_outgoing) > SSH_BUFFERING_LIMIT)
    {
      SSH_DEBUG(5, ("BLOCKING: up_outgoing too big."));
      tr->read_has_blocked = TRUE;
      return FALSE;
    }

  /* Try to read a packet. */
  packet = ssh_tr_input_packet(tr);
  if (!packet)
    return FALSE;

  /* Get the packet type, without actually removing it from the packet. */
  packet_type = ssh_tr_peek_packet_type(packet);

  /* Process packets belonging to services by passing then up. */
  if (packet_type >= SSH_FIRST_SERVICE_PACKET)
    {
      /* Pass the packet upwards. */
      ssh_tr_up_send(tr, SSH_CROSS_PACKET,
                     ssh_buffer_ptr(packet), ssh_buffer_len(packet));
      ssh_buffer_free(packet);
      return TRUE;
    }

  /* Kex packets are illegal here, as are many of the known packets. */
  if (packet_type >= SSH_FIRST_KEX_PACKET ||
      packet_type == SSH_MSG_NEWKEYS ||
      packet_type == SSH_MSG_SERVICE_REQUEST ||
      packet_type == SSH_MSG_SERVICE_ACCEPT)
    {
      /* We should never receive kex packets in this state. */
      ssh_tr_up_disconnect(tr, TRUE, TRUE,
                           SSH_DISCONNECT_PROTOCOL_ERROR,
                           "Protocol error: packet %d in interactive",
                           packet_type);
      ssh_buffer_free(packet);
      return FALSE;
    }

  /* Currently, the only supported packet here is KEXINIT, which will
     initiate rekey. */
  if (packet_type == SSH_MSG_KEXINIT)
    {
      /* Start rekey.  This will cause the packet to be freed. */
      ssh_tr_input_start_rekey(tr, packet);
      return TRUE;
    }

  /* This is an unknown packet. */
  ssh_buffer_free(packet);
  ssh_buffer_init(&buffer);
  ssh_bufaux_put_char(&buffer, SSH_MSG_UNIMPLEMENTED);
  SSH_PUT_32BIT(seq_buf, tr->incoming_sequence_number - 1);
  ssh_xbuffer_append(&buffer, seq_buf, 4);
  ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer),
                     ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);
  return TRUE;
}

/* Process input from the connection. */

void ssh_tr_process_input(SshTransportCommon tr)
{
  Boolean ok = TRUE;

  SSH_DEBUG(9, ("Processing input from the connection."));

  while (ok)
    {
      ok = FALSE;
      switch (tr->received_state)
        {
        case RECEIVED_NOTHING:
          ok = ssh_tr_input_version(tr);
          break;

        case RECEIVED_VERSION:
          ok = ssh_tr_input_kexinit(tr);
          break;

        case RECEIVED_KEXINIT:
          if ((tr->server ? tr->kex->server_input_kex1
               : tr->kex->client_input_kex1) != NULL_FNPTR)
            ok = ssh_tr_input_kex1(tr);
          else
            ssh_tr_input_kex2(tr);
          break;

        case RECEIVED_KEX1_IGNORED:
          ok = ssh_tr_input_kex1(tr);
          break;

        case RECEIVED_KEX1_FINAL:
          ssh_tr_input_kex2(tr);
          break;

        case RECEIVED_KEX2:
          ok = ssh_tr_input_newkeys(tr);
          break;

        case RECEIVED_KEY_CHECK:
          /* We get moved to the next state after we receive the key check
             response. */
          SSH_DEBUG(5, ("BLOCKED: wait key check"));
          ok = FALSE;
          break;

        case RECEIVED_NEWKEYS:
          /* If doing rekey, we can now proceed with receiving normal data. */
          if (tr->doing_rekey)
            {
              /* Send a SSH_CROSS_ALGORITHMS packet upwards. */
              ssh_tr_up_send_algorithms(tr);

              /* Update state. */
              tr->received_state = RECEIVED_INTERACTIVE;
              tr->rekey_request_sent = FALSE;
              /* If we have already sent NEWKEYS, rekey operation is
                 complete. */
              if (tr->sent_state == SENT_INTERACTIVE)
                tr->doing_rekey = FALSE;
              tr->successful_rekeys++;
              ok = TRUE;
              break;
            }

          /* If server, wait for service request; if client, wait for service
             accept.  These will update state when appropriate. */
          if (tr->server)
            ok = ssh_tr_input_service_request(tr);
          else
            ok = ssh_tr_input_service_accept(tr);
          break;

        case RECEIVED_SERVICE_REQUEST:
          /* We are waiting for service accept or disconnect from up.
             We will be automatically advanced when we receive it. */
          SSH_DEBUG(5, ("BLOCKING: up service accept wait"));
          tr->read_has_blocked = TRUE;
          ok = FALSE;
          break;

        case RECEIVED_INTERACTIVE:
          /* We are passing packets to the service. */
          ok = ssh_tr_input_interactive(tr);
          break;

        case RECEIVED_DEAD:
          /* We will not do anything more. */
          break;

        default:
          ssh_fatal("ssh_tr_process_input: unknown received state %d",
                    (int)tr->received_state);
        }
    }
}

/* Sends a service request packet.  This does not update state. */

void ssh_tr_send_service_request(SshTransportCommon tr)
{
  SshBufferStruct buffer;

  SSH_DEBUG(4, ("Sending service request for \"%s\".", tr->service_name));

  ssh_buffer_init(&buffer);
  ssh_bufaux_put_char(&buffer, SSH_MSG_SERVICE_REQUEST);
  ssh_bufaux_put_uint32_string(&buffer, tr->service_name,
                               strlen(tr->service_name));
  ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer),
                     ssh_buffer_len(&buffer));
  ssh_buffer_uninit(&buffer);
}

/* Send pending output to the connection. */

void ssh_tr_process_output(SshTransportCommon tr)
{
  Boolean ok = TRUE;

  SSH_DEBUG(9, ("Sending pending output to the connection."));

  while (ok)
    {
      /* If we have pending data to output, process it first.  Note that
         this is processed also in SENT_DEAD state. */
      if (ssh_buffer_len(&tr->outgoing) > 0)
        {
          ok = ssh_tr_output_outgoing(tr);
          continue;
        }

      switch (tr->sent_state)
        {
        case SENT_NOTHING:
          /* This is technically against the draft. */
          /* Don't send anything before we receive the server's
             version string. */
          /* The reason for this kludge is that there existed a race
             condition. Sometimes ssh2-client executed the ssh1-client
             (=the version_callback got called), sometimes not,
             depending on whether it managed to process the server's
             version string before the server had closed the
             connection, and the stream got destroyed. It wasn't nice,
             so now we get to exec it in any case. Furthermore, this
             bug wasn't universal, and I found it when using my
             Linux-system. */
          if (!tr->server && tr->version_compatibility &&
              tr->received_state < RECEIVED_VERSION )
            {
              /* We're in compatibility mode, we're client, and we
                 haven't received server's version string yet. */
              ok = FALSE;
              break;
            }

          /* Append version number to outgoing data. */
          ssh_xbuffer_append(&tr->outgoing, (unsigned char *) tr->own_version,
                             strlen(tr->own_version));

          /* No CR on compat mode */
          if (strncmp("SSH-1.", tr->own_version, 6) != 0)
            ssh_xbuffer_append(&tr->outgoing, (unsigned char *) "\r", 1);
          ssh_xbuffer_append(&tr->outgoing, (unsigned char *) "\n", 1);
          tr->sent_state = SENT_VERSION;
          /* We loop again... */
          break;

        case SENT_VERSION:
          /* Do not send further data after version until we have received
             client version. */
          if (tr->server && tr->version_compatibility &&
              tr->received_state < RECEIVED_VERSION)
            {
              ok = FALSE;
              break;
            }

          tr->sent_state = SENT_KEXINIT;

          /* Construct and save a kexinit packet. */
          ssh_tr_output_kexinit(tr);
          /* We loop again... */
          break;

        case SENT_KEXINIT:
        case SENT_KEX1_FINAL:
          /* We must wait for received KEXINIT packet before we can
             continue.  Receiving KEXINIT will automatically trigger sending
             a new KEX1 packet if appropriate.  Thus, that need not be
             handled here.  Basically we just linger on here until it is time
             to send KEX2.  However, sending KEX2 is automatically triggered
             by receiving KEX1.  Thus, we need to wait until we have received
             KEX2.  But, NEWKEYS is automatically sent in response to receiving
             it, and thus we must wait here until NEWKEYS have been received.
             Simple, eh? */
          ok = FALSE;  /* We are automatically advanced when we have sent
                          NEWKEYS. */
          break;

        case SENT_NEWKEYS:
          /* If rekey, continue normal processing for output now that we have
             sent NEWKEYS. */
          if (tr->doing_rekey)
            {
              tr->sent_state = SENT_INTERACTIVE;
              tr->rekey_request_sent = FALSE;
              /* If we have already received NEWKEYS, rekey operation
                 is complete. */
              if (tr->received_state == RECEIVED_INTERACTIVE)
                tr->doing_rekey = FALSE;
              break; /* We loop again... */
            }

          /* If server, we are automatically advanced when we receive service
             request.  We just sleep here until then. */
          if (tr->server)
            {
              ok = FALSE;
              break;
            }

          /* We are the client, and this is the initial key exchange.
             Send service request. */
          ssh_tr_send_service_request(tr);
          tr->sent_state = SENT_SERVICE_REQUEST;
          break;  /* We loop again... */

        case SENT_SERVICE_REQUEST:
          /* (We only get here as client.)  If not using encryption, do
             not proceed until we have received service accept. */
          SSH_ASSERT(!tr->server);
          if (strcmp(tr->c_to_s.cipher_name, "none") == 0 &&
              tr->received_state < RECEIVED_INTERACTIVE)
            {
              ok = FALSE;
              break;
            }
          /* We can proceed. */
          tr->sent_state = SENT_INTERACTIVE;
          /* Note: SSH_CROSS_STARTUP isn't sent until we receive service
             accept. */
          break; /* We loop again... */

        case SENT_INTERACTIVE:
          /* In this state, we don't really send anything automatically.
             Sending user data is handled in ssh_tr_up_write, and by
             ssh_tr_output_outgoing above. */
          ok = FALSE;
          break;

        case SENT_DEAD:
          /* We will not do anything more. */
          ok = FALSE;
          break;

        default:
          ssh_fatal("ssh_tr_process_output: unknown state %d",
                    (int)tr->sent_state);
        }
    }
}

/* Called whenever input/output is available on the stream, or a special
   condition occurs. */

void ssh_tr_callback(SshStreamNotification notification,
                     void *context)
{
  SshTransportCommon tr = context;

  /* Just in case we would get queued events during destruction... */
  if (tr->connection == NULL)
    return;

  /* Process the notification. */
  switch (notification)
    {
    case SSH_STREAM_INPUT_AVAILABLE:
      ssh_tr_process_input(tr);
      break;

    case SSH_STREAM_CAN_OUTPUT:
      ssh_tr_process_output(tr);
      break;

    case SSH_STREAM_DISCONNECTED:
      ssh_tr_up_disconnect(tr, TRUE, FALSE,
                           SSH_DISCONNECT_CONNECTION_LOST,
                           "The connection has been lost.");
      break;

    default:
      ssh_fatal("ssh_tr_callback: unknown notification %d",
                (int)notification);
    }
}

/* Process a cross-layer packet received from up_stream. */

void ssh_tr_process_up_incoming_packet(SshTransportCommon tr,
                                       unsigned int packet_type,
                                       const unsigned char *payload,
                                       size_t payload_len)
{
  SshBufferStruct buffer;
  Boolean always_display;
  unsigned int tr_packet_type;
  char *ciphers_c_to_s, *ciphers_s_to_c, *macs_c_to_s, *macs_s_to_c,
    *compressions_c_to_s, *compressions_s_to_c, *host_key_algorithms,
    *service_name;
  unsigned char *msg, *msg_lang;
  SshUInt32 reason_code;

  SSH_DEBUG(9, ("Processing packet from the application. (type: %d)",
                packet_type));

  switch (packet_type)
    {
    case SSH_CROSS_PACKET:
      /* We have already done flow control; no need to do it here.  Just
         pass the packet down.  However, SSH_MSG_DISCONNECT packets get
         special handling; they cause SSH_CROSS_DISCONNECT to be
         relayed up. */
      tr_packet_type = payload[0];

      /* If waiting for service accept, don't accept anything else. */
      if (tr->received_state == RECEIVED_SERVICE_REQUEST &&
          tr_packet_type != SSH_MSG_DISCONNECT)
        ssh_fatal("ssh_tr_process_up_incoming_packet: expected "
                  "SERVICE_ACCEPT or DISCONNECT");

      if (tr_packet_type == SSH_MSG_DISCONNECT)
        ssh_fatal("ssh_tr_process_up_incoming_packet: "
                  "received SSH_MSG_DISCONNECT.  The interface has changed; "
                  "these now need to be sent as SSH_CROSS_DISCONNECT "
                  "cross-layer packets.");

      if (tr_packet_type < SSH_FIRST_SERVICE_PACKET &&
          tr_packet_type != SSH_MSG_DISCONNECT &&
          tr_packet_type != SSH_MSG_IGNORE)
        {
          ssh_tr_up_disconnect(tr, TRUE, TRUE,
                               SSH_DISCONNECT_PROTOCOL_ERROR,
                               "Protocol error: service sending tr packet %d",
                               tr_packet_type);
          return;
        }

      /* Padding auth packets to specified increment. */
      if (tr_packet_type >= SSH_FIRST_USERAUTH_PACKET &&
          tr_packet_type <= SSH_LAST_USERAUTH_PACKET)
        {
          SshBuffer new_packet = ssh_xbuffer_allocate();
#define AUTH_PADDING_WIDTH 1024
          long ignore_len = AUTH_PADDING_WIDTH -
              payload_len % AUTH_PADDING_WIDTH;
          long i = 0L;
          unsigned char *datap;

          SSH_DEBUG(6, ("Padding auth packet with an SSH_MSG_IGNORE packet "
                        "with %d bytes of payload (payload_len = %d).",
                        ignore_len, payload_len));

          ssh_encode_buffer(new_packet,
                            SSH_FORMAT_CHAR, (unsigned int) SSH_MSG_IGNORE,
                            SSH_FORMAT_UINT32, ignore_len,
                            SSH_FORMAT_END);

          ssh_xbuffer_append_space(new_packet, &datap, ignore_len);

          for (i = 0; i < ignore_len; i++)
            datap[i] = tr_get_padding_byte(tr);

          ssh_tr_encode_packet(tr, ssh_buffer_ptr(new_packet),
                               ssh_buffer_len(new_packet));
          ssh_buffer_free(new_packet);
        }

      /* Send the packet to the connection. */
      ssh_tr_send_packet(tr, payload, payload_len);
      break;

    case SSH_CROSS_DISCONNECT:
      /* Received a disconnect packet from up.  We should send it to the
         other side, relay it back up, and disconnect. */
      if (ssh_decode_array(payload, payload_len,
                           SSH_FORMAT_BOOLEAN, NULL,
                           SSH_FORMAT_UINT32, &reason_code,
                           SSH_FORMAT_UINT32_STR, &msg, NULL,
                           SSH_FORMAT_UINT32_STR, &msg_lang, NULL,
                           SSH_FORMAT_END) == 0)
        ssh_fatal("ssh_tr_process_up_incoming_packet: bad DISCONNECT");

      SSH_DEBUG(4, ("Received disconnect packet from application."));

      /* Send the DISCONNECT packet back up. */
      ssh_tr_up_send(tr, SSH_CROSS_DISCONNECT,
                     payload, payload_len);

      /* Compatibility code with older versions. */
      if (tr->ssh_old_deprecated_disconnect_codes_compat)
        {
          if (reason_code > SSH_DISCONNECT_BY_APPLICATION ||
              reason_code == SSH_DISCONNECT_RESERVED)
            reason_code = SSH_DISCONNECT_BY_APPLICATION;
        }

      /* Send the disconnect packet to the connection. */
      ssh_buffer_init(&buffer);
      ssh_encode_buffer(&buffer,
                        SSH_FORMAT_CHAR, (unsigned int) SSH_MSG_DISCONNECT,
                        SSH_FORMAT_UINT32, reason_code,
                        SSH_FORMAT_UINT32_STR, msg,
                        strlen((char *) msg),
                        SSH_FORMAT_UINT32_STR, msg_lang,
                        strlen((char *) msg_lang),
                        SSH_FORMAT_END);
      ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
      ssh_buffer_uninit(&buffer);

      /* Prepare for shutdown. */
      tr->up_outgoing_eof = TRUE;
      tr->received_state = RECEIVED_DEAD;
      tr->sent_state = SENT_DEAD;
      tr->outgoing_eof = TRUE;

      ssh_xfree(msg);
      ssh_xfree(msg_lang);

      ssh_tr_output_outgoing(tr);
      ssh_tr_up_signal_input(tr);
      return;

    case SSH_CROSS_DEBUG:
      /* Received a debug packet from up.  We should send a debug
         packet to the other side. */
      if (ssh_decode_array(payload, payload_len,
                           SSH_FORMAT_BOOLEAN, &always_display,
                           SSH_FORMAT_UINT32_STR, &msg, NULL,
                           SSH_FORMAT_UINT32_STR, &msg_lang, NULL,
                           SSH_FORMAT_END) == 0)
        ssh_fatal("ssh_tr_process_up_incoming_packet: bad DEBUG");

      /* Send the debug packet to the connection. */
      ssh_buffer_init(&buffer);
      ssh_encode_buffer(&buffer,
                        SSH_FORMAT_CHAR, (unsigned int) SSH_MSG_DEBUG,
                        SSH_FORMAT_BOOLEAN, always_display,
                        SSH_FORMAT_UINT32_STR, msg,
                        strlen((char *) msg),
                        SSH_FORMAT_UINT32_STR, msg_lang,
                        strlen((char *) msg_lang),
                        SSH_FORMAT_END);
      ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer), ssh_buffer_len(&buffer));
      ssh_buffer_uninit(&buffer);

      ssh_xfree(msg);
      ssh_xfree(msg_lang);

      break;

    case SSH_CROSS_REKEY_REQUEST:
      if (tr->received_state == RECEIVED_SERVICE_REQUEST)
        ssh_fatal("ssh_tr_process_up_incoming_packet: expected "
                  "SERVICE_ACCEPT or DISCONNECT");

      if (tr->doing_rekey == TRUE)
        {
          /* This shouldn't usually happen, but in special cases when
             the timing is just right, this check is necessary. ex:
             application has sent REKEY_REQUEST downstream. Transport
             layer gets a KEXINIT package from network before reading
             stuff from upstream. Transport notifies application of a
             beginning rekey operation, BUT applications request is
             already waiting in transports upstream. Now, transport
             reads upstream, and bang! gets the rekey request from the
             application in the middle of rekeying. Not good, but I
             don't know how to avoid that. So, to avoid ASSERTing
             below, this check is needed. */
          SSH_TRACE(2, ("Already doing rekey, ignoring application "
                        "request..."));
          break;
        }

      /* Get new algorithms from the payload. */
      if (ssh_decode_array(payload, payload_len,
                           SSH_FORMAT_UINT32_STR, &ciphers_c_to_s, NULL,
                           SSH_FORMAT_UINT32_STR, &ciphers_s_to_c, NULL,
                           SSH_FORMAT_UINT32_STR, &macs_c_to_s, NULL,
                           SSH_FORMAT_UINT32_STR, &macs_s_to_c, NULL,
                           SSH_FORMAT_UINT32_STR, &compressions_c_to_s, NULL,
                           SSH_FORMAT_UINT32_STR, &compressions_s_to_c, NULL,
                           SSH_FORMAT_UINT32_STR, &host_key_algorithms, NULL,
                           SSH_FORMAT_END) == 0)
        ssh_fatal("ssh_tr_process_up_incoming_packet: bad REKEY_REQUEST");

      /* Start rekey. */
      SSH_ASSERT(tr->sent_state == SENT_INTERACTIVE);
      SSH_ASSERT(tr->received_state == RECEIVED_INTERACTIVE);

      /* Notify application, that the rekey operation started. */
      if (tr->rekey_notify_callback)
        (*tr->rekey_notify_callback)(tr->rekey_notify_context);

      ssh_tr_kex_cleanup(tr);
      tr->doing_rekey = TRUE;
      tr->rekey_request_sent = TRUE;
      tr->sent_state = SENT_VERSION;
      ssh_tr_output_kexinit_explicit(tr,
                                     ciphers_c_to_s,
                                     ciphers_s_to_c,
                                     macs_c_to_s,
                                     macs_s_to_c,
                                     compressions_c_to_s,
                                     compressions_s_to_c,
                                     host_key_algorithms);
      ssh_xfree(ciphers_c_to_s);
      ssh_xfree(ciphers_s_to_c);
      ssh_xfree(macs_c_to_s);
      ssh_xfree(macs_s_to_c);
      ssh_xfree(compressions_c_to_s);
      ssh_xfree(compressions_s_to_c);
      ssh_xfree(host_key_algorithms);

      tr->sent_state = SENT_KEXINIT;
      break;

    case SSH_CROSS_SERVICE_ACCEPT:
      /* The service was accepted.  Send a service accept. */
      if (tr->ssh_old_msg_service_accept_bug_compat)
        {
          ssh_tr_send_simple_packet(tr, SSH_MSG_SERVICE_ACCEPT);
        }
      else
        {
          if (ssh_decode_array(payload, payload_len,
                               SSH_FORMAT_UINT32_STR, &service_name, NULL,
                               SSH_FORMAT_END) == 0)
            ssh_fatal("ssh_tr_process_up_incoming_packet: bad SERVICE_ACCEPT");

          ssh_buffer_init(&buffer);
          ssh_encode_buffer(&buffer,
                            SSH_FORMAT_CHAR,
                            (unsigned int) SSH_MSG_SERVICE_ACCEPT,
                            SSH_FORMAT_UINT32_STR, service_name,
                            strlen((char *)service_name),
                            SSH_FORMAT_END);
          ssh_tr_send_packet(tr, ssh_buffer_ptr(&buffer),
                             ssh_buffer_len(&buffer));
          ssh_buffer_uninit(&buffer);

          ssh_xfree(service_name);

        }

      /* Send a SSH_CROSS_STARTUP packet upwards. */
      ssh_tr_up_send_startup(tr);

      /* Send a SSH_CROSS_ALGORITHMS packet upwards. */
      ssh_tr_up_send_algorithms(tr);

      /* Update state. */
      tr->received_state = RECEIVED_INTERACTIVE;
      tr->sent_state = SENT_INTERACTIVE;

      ssh_tr_process_output(tr);
      ssh_tr_wake_up_input(tr);
      break;

    default:
      ssh_fatal("ssh_tr_process_up_incoming_packet: unexpected packet %d",
                packet_type);
    }
}

/* Up_stream read operation.  This is called by the generic streams code. */

int ssh_tr_up_read(void *context, unsigned char *buf, size_t size)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(10, ("Application reading at most %ld bytes from transport "
                 "stream.", size));

  /* We cannot read more data than is available in the buffer. */
  if (size > ssh_buffer_len(&tr->up_outgoing))
    size = ssh_buffer_len(&tr->up_outgoing);

  /* If no data available, return EOF or error. */
  if (size == 0)
    {
      if (tr->up_outgoing_eof)
        {
          return 0;
        }
      else
        {
          tr->up_read_blocked = TRUE;
          return -1;
        }
    }

  /* Copy data to user buffer. */
  memcpy(buf, ssh_buffer_ptr(&tr->up_outgoing), size);
  ssh_buffer_consume(&tr->up_outgoing, size);

  /* Make sure we are receiving more input from the network. */
  ssh_tr_wake_up_input(tr);

  return size;
}

/* Up_stream write operation.  This is called by the generic streams code
   WHEN STUFF IS COMING DOWN TO THE TRANSPORT LAYER */

int ssh_tr_up_write(void *context, const unsigned char *buf, size_t size)
{
  SshTransportCommon tr = context;
  size_t offset, payload_len, len;
  const unsigned char *ucp;

  SSH_DEBUG(10, ("Application writing %ld bytes to transport stream.",
                 size));

  if (tr->sent_state != SENT_INTERACTIVE &&
      tr->received_state != RECEIVED_SERVICE_REQUEST)
    {
      tr->up_write_blocked = TRUE;
      return -1;
    }

  if (tr->outgoing_eof)
    return 0;

  offset = 0;

  if (ssh_buffer_len(&tr->up_incoming) > 0)
    goto partial;

 normal:

  while (ssh_buffer_len(&tr->outgoing) <
         XMALLOC_MAX_SIZE - SSH_MAX_TOTAL_PACKET_LENGTH - SSH_CONTROL_RESERVE
         && ssh_buffer_len(&tr->outgoing) < SSH_BUFFERING_LIMIT)
    {
      /* We only accept data from up in interactive state, and not after having
         already scheduled eof to connection. */
      if (tr->sent_state != SENT_INTERACTIVE &&
          tr->received_state != RECEIVED_SERVICE_REQUEST)
        break;

      if (tr->outgoing_eof)
        break;

      /* If we have processed all data, return now. */
      if (offset == size)
        return offset;

      /* If only partial packet available, do special processing. */
      if (size - offset < 4)
        goto partial; /* Need partial packet processing. */
      payload_len = SSH_GET_32BIT(buf + offset);
      if (size - offset < 4 + payload_len)
        goto partial; /* Need partial packet processing. */

      /* The entire packet is available; process it now. */
      ssh_tr_process_up_incoming_packet(tr, buf[offset + 4],
                                        buf + offset + 5, payload_len - 1);
      offset += 4 + payload_len;
    }
  /* We cannot process more data now. */
  if (offset > 0)
    return offset;
  tr->up_write_blocked = TRUE;
  return -1;

 partial:
  /* Process partial packet. */
  len = ssh_buffer_len(&tr->up_incoming);
  if (len < 4)
    {
      len = 4 - len;
      if (size - offset < len)
        len = size - offset;
      ssh_xbuffer_append(&tr->up_incoming, buf + offset, len);
      offset += len;
    }
  if (ssh_buffer_len(&tr->up_incoming) < 4)
    return offset;
  ucp = ssh_buffer_ptr(&tr->up_incoming);
  payload_len = SSH_GET_32BIT(ucp);
  len = 4 + payload_len - ssh_buffer_len(&tr->up_incoming);
  if (len > size - offset)
    len = size - offset;
  ssh_xbuffer_append(&tr->up_incoming, buf + offset, len);
  offset += len;
  if (ssh_buffer_len(&tr->up_incoming) < 4 + payload_len)
    return offset;

  /* The entire packet is now in buffer. */
  ucp = ssh_buffer_ptr(&tr->up_incoming);
  ssh_tr_process_up_incoming_packet(tr, ucp[4], ucp + 5, payload_len - 1);

  /* Clear the incoming partial packet buffer and resume normal processing. */
  ssh_buffer_clear(&tr->up_incoming);
  goto normal;
}

/* Indicates that the application will not write anymore.  We will basically
   just close the connection. */

void ssh_tr_up_output_eof(void *context)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(5, ("Application put EOF to transport stream."));

  if (tr->outgoing_eof)
    return;

  tr->outgoing_eof = TRUE;
  ssh_tr_output_outgoing(tr);
}

/* Sets the callback used to notify the application.  The callback may be
   NULL. */

void ssh_tr_up_set_callback(void *context,
                            SshStreamCallback up_callback,
                            void *up_context)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(10, ("Application changing transport stream callback."));

  /* Set the callback. */
  tr->up_callback = up_callback;
  tr->up_context = up_context;

  /* Cause the application callback to be called for both reading and
     writing (from the bottom of the event loop). */
  ssh_tr_up_signal_input(tr);
  ssh_tr_up_signal_output(tr);
}

/* Destroys the stream context. */

void ssh_tr_up_destroy(void *context)
{
  SshTransportCommon tr = context;

  SSH_DEBUG(4, ("Destroying transport stream."));

  tr->up_stream = NULL;
  tr->up_callback = NULL_FNPTR;
  tr->received_state = RECEIVED_DEAD;
  tr->sent_state = SENT_DEAD;

  if (ssh_buffer_len(&tr->outgoing) == 0 ||
      tr->connection == NULL)
    ssh_tr_destroy_now(tr);
}

/* Method table for the up_stream. */

SshStreamMethodsStruct ssh_tr_methods =
  {
    ssh_tr_up_read,
    ssh_tr_up_write,
    ssh_tr_up_output_eof,
    ssh_tr_up_set_callback,
    ssh_tr_up_destroy
  };

/* Creates the SshTransportCommon object, and performs initializations that
   are common to client and server.  Either client or server initialization
   should be performed after this call. */

SshTransportCommon ssh_tr_create(SshStream connection,
                                 Boolean server,
                                 Boolean compatibility,
                                 Boolean fake_old_version,
                                 const char *application_version,
                                 SshTransportParams params)
{
  SshTransportCommon tr;
  char buf[256];

  tr = ssh_xcalloc(1, sizeof(*tr));

  /* Initialize general state. */
  tr->server = server;
  tr->version_compatibility = compatibility;
  tr->doing_rekey = FALSE;
  tr->rekey_request_sent = FALSE;
  tr->read_has_blocked = FALSE;
  tr->sent_state = SENT_NOTHING;
  tr->received_state = RECEIVED_NOTHING;
  tr->connection = connection;
  tr->params = params;

  /* Initialize packet sequence numbers. */
  tr->incoming_sequence_number = 0;
  tr->outgoing_sequence_number = 0;

  tr->outgoing_eof = FALSE;

  /* Initialize incoming/outgoing buffers. */
  ssh_buffer_init(&tr->outgoing);
  tr->incoming_packet = NULL;
  ssh_buffer_init(&tr->up_outgoing);
  ssh_buffer_init(&tr->up_incoming);
  tr->up_write_blocked = FALSE;
  tr->up_read_blocked = FALSE;

  /* Initialize guessed algorithms. */
  tr->guessed_kex = ssh_snlist_get_name(params->kex_algorithms);
  tr->guessed_host_key = ssh_snlist_get_name(params->host_key_algorithms);
  tr->send_kex1_immediately = TRUE;

  tr->kex = ssh_kex_lookup(tr->guessed_kex);
  if (!tr->kex)
    ssh_fatal("ssh_tr_create: guessed kex '%.100s' not found",
              tr->guessed_kex);

  /* Initialize current algorithms. */
  tr->kex_name = ssh_xstrdup(tr->guessed_kex);
  tr->host_key_name = ssh_xstrdup(tr->guessed_host_key);
  tr->host_key_names = ssh_xstrdup(params->host_key_algorithms);
  tr->c_to_s.cipher_name = ssh_xstrdup("none");
  tr->c_to_s.mac_name = ssh_xstrdup("none");
  tr->c_to_s.compression_name = ssh_xstrdup("none");
  tr->s_to_c.cipher_name = ssh_xstrdup("none");
  tr->s_to_c.mac_name = ssh_xstrdup("none");
  tr->s_to_c.compression_name = ssh_xstrdup("none");
  if (tr->server)
    {
      ssh_tr_set_keys(tr, &tr->c_to_s, &tr->incoming_granularity, FALSE,
                      &tr->incoming_cipher, &tr->incoming_mac,
                      &tr->compression_incoming);
      ssh_tr_set_keys(tr, &tr->s_to_c, &tr->outgoing_granularity, TRUE,
                      &tr->outgoing_cipher, &tr->outgoing_mac,
                      &tr->compression_outgoing);
    }
  else
    {
      ssh_tr_set_keys(tr, &tr->s_to_c, &tr->incoming_granularity, FALSE,
                      &tr->incoming_cipher, &tr->incoming_mac,
                      &tr->compression_incoming);
      ssh_tr_set_keys(tr, &tr->c_to_s, &tr->outgoing_granularity, TRUE,
                      &tr->outgoing_cipher, &tr->outgoing_mac,
                      &tr->compression_outgoing);
    }

  tr->compression_buffer = ssh_xbuffer_allocate();
  tr->compressed_incoming_bytes = 0;
  tr->uncompressed_incoming_bytes = 0;
  tr->compressed_outgoing_bytes = 0;
  tr->uncompressed_outgoing_bytes = 0;

  SSH_ASSERT(sizeof(buf) >= 256);
  ssh_snprintf(buf, 256 - 3,
               fake_old_version
               ? SSH_VERSION_STRING_COMPAT
               : SSH_VERSION_STRING,
               application_version);
  tr->own_version = ssh_xstrdup(buf);
  SSH_DEBUG(2, ("My version: %s", tr->own_version));

  tr->public_host_key = NULL;
  tr->public_host_key_blob = NULL;
  tr->host_keys_ctx = NULL;
  
  tr->session_identifier_len = 0;
  tr->exchange_hash_len = 0;

  tr->successful_rekeys = 0L;

  /* All compatibility flags are initially FALSE, so no initialization
     is needed because of the ssh_xcalloc(). */
  
  /* initialize the key exchange parameters */

  ssh_mp_init(tr->dh_p);
  ssh_mp_init(tr->dh_g);
  ssh_mp_init(tr->dh_e);
  ssh_mp_init(tr->dh_f);
  ssh_mp_init(tr->dh_k);
  ssh_mp_init(tr->dh_secret);

  return tr;
}

/* Finalizes the creation of the transport layer protocol.  Wraps it into
   a stream, and returns the stream.  The lower-level object should not
   be accessed after this call; the object will be automatically destroyed
   when the stream is destroyed. */

SshStream ssh_tr_create_final(SshTransportCommon tr)
{
  SshStream up_stream;
  up_stream = ssh_stream_create(&ssh_tr_methods, (void *)tr);
  tr->up_stream = up_stream;
  ssh_stream_set_callback(tr->connection, ssh_tr_callback, (void *)tr);
  return up_stream;
}
