/*

  sshcommon.c

  Authors:
        Tatu Ylonen <ylo@ssh.com>
        Markku-Juhani Saarinen <mjos@ssh.com>
        Timo J. Rinne <tri@ssh.com>
        Sami Lehtinen <sjl@ssh.com>

  Copyright (C) 1997-2002 SSH Communications Security Corp, Helsinki, Finland
  All rights reserved.

  This file implements the core of the SSH2 channel protocols.  This interface
  is shared by server and client code.  Most of the real stuff is in
  channel type specific modules (sshch*.c and sshch*.h).  This file
  is mostly just glue to put them together.

  To define a new channel type, write sshch*.h and sshch*.c files for
  it, and add it to the tables in this module.  Use existing channel
  types as a model.

*/

#include "ssh2includes.h"
#include "sshencode.h"
#include "sshmsgs.h"
#include "sshuser.h"
#include "sshcommon.h"
#include "sshtimeouts.h"
#include "sshnameserver.h"
#include "sshadt_list.h"

#ifdef SSH_CHANNEL_SESSION
#include "sshchsession.h"
#endif /* SSH_CHANNEL_SESSION */

#ifdef SSH_CHANNEL_X11
#include "sshchx11.h"
#endif /* SSH_CHANNEL_X11 */

#ifdef SSH_CHANNEL_AGENT
#include "sshchagent.h"
#endif /* SSH_CHANNEL_AGENT */
#ifdef SSH_CHANNEL_SSH1_AGENT
#include "sshchssh1agent.h"
#endif /* SSH_CHANNEL_SSH1_AGENT */

#ifdef SSH_CHANNEL_TCPFWD
#include "sshchtcpfwd.h"
#endif /* SSH_CHANNEL_TCPFWD */

#define SSH_DEBUG_MODULE "Ssh2Common"

/* This array completely defines all channel types and global requests
   supported by the system.  To add a new channel, write the appropriate
   header file and implementation, include the header into this file, and
   add the channel type into this array. */

struct {
  const char *name;



  SshConnChannelOpenProc open_proc;

  SshChannelTypeCreateProc type_create_proc;
  SshChannelTypeDestroyProc type_destroy_proc;



  SshConnGlobalRequestStruct global_requests[10];

} ssh_channel_types[] =
{

#ifdef SSH_CHANNEL_SESSION
  /* Interactive session channels. */
  { "session", ssh_channel_session_open, NULL_FNPTR, NULL_FNPTR },
#endif /* SSH_CHANNEL_SESSION */

#ifdef SSH_CHANNEL_X11
  /* X11 forwarding channels. */
  { "x11", ssh_channel_x11_open, ssh_channel_x11_create,
    ssh_channel_x11_destroy },
#endif /* SSH_CHANNEL_X11 */

#ifdef SSH_CHANNEL_AGENT
  /* Authentication agent forwarding channels. */
  { "auth-agent", ssh_channel_agent_open, ssh_channel_agent_create,
    ssh_channel_agent_destroy },
#endif /* SSH_CHANNEL_AGENT */

#ifdef SSH_CHANNEL_SSH1_AGENT
  /* Ssh1 authentication agent forwarding channels. */
  { "auth-ssh1-agent", ssh_channel_ssh1_agent_open,
    ssh_channel_ssh1_agent_create, ssh_channel_ssh1_agent_destroy },
#endif /* SSH_CHANNEL_SSH1_AGENT */

#ifdef SSH_CHANNEL_TCPFWD
#ifndef SSH_WIN_CLIENT
  /* Forwarded TCP/IP channels. */
    /* on client: receives a remote forward connection and connects
       to destination */
  { "forwarded-tcpip", ssh_channel_ftcp_open_request, ssh_channel_ftcp_create,
    ssh_channel_ftcp_destroy,
  {
    /* creates a listener on the server */
    { "tcpip-forward", ssh_channel_remote_tcp_forward_request },
    /* destroyes a listener on the server */
    { "cancel-tcpip-forward", ssh_channel_tcp_forward_cancel },
    { NULL, NULL_FNPTR } } },
    /* on server: receives a local forward and connects to destination */
  { "direct-tcpip", ssh_channel_dtcp_open_request, ssh_channel_dtcp_create,
    ssh_channel_dtcp_destroy },
#else /* !SSH_WIN_CLIENT */
  { "forwarded-tcpip", ssh_channel_ftcp_open_request, ssh_channel_ftcp_create,
    ssh_channel_ftcp_destroy },
    /* Client doesn't accept forwards from the server. Hence the NULL below.*/
  { "direct-tcpip", NULL, ssh_channel_dtcp_create,
    ssh_channel_dtcp_destroy },
#endif /* !SSH_WIN_CLIENT */
#endif /* SSH_CHANNEL_TCPFWD */

  { NULL },
};

/* Clean-up hooks. */
typedef struct CleanUpRec {
  SshCommonCleanUpCB clean_up;
  void *context;
} CleanUpStruct, *CleanUp;

void ssh_common_destroy_active_forward(SshTcpForwardActive fwd)
{
  if (fwd->handle)
    ssh_operation_abort(fwd->handle);
  ssh_xfree(fwd->connect_from_host);
  ssh_xfree(fwd->port);
  ssh_xfree(fwd->connect_to_host);
  ssh_xfree(fwd->connect_to_port);
  ssh_xfree(fwd);
}

/* For ADT. */
static void destroy_forward_rec(void *obj, void *context)
{
  ssh_common_destroy_active_forward((SshTcpForwardActive)obj);
}

/* Timeout to destroy the common structure from the bottom of the event
   loop. */

void common_destroy_timeout(void *ctx)
{
  SshCommon common = (SshCommon)ctx;

  ssh_common_destroy(common);
}

/* Processes a disconnect message received from the connection protocol.
   Currently, a DISCONNECT message is sent if the protocol terminates. */

void ssh_common_disconnect(int reason, Boolean locally_generated,
                           const char *msg, void *context)
{
  SshCommon common = (SshCommon)context;

  SSH_DEBUG(2, ("DISCONNECT received: %s", msg));

  /* Log the disconnect in the system log. */
  ssh_log_event(common->config->log_facility, SSH_LOG_INFORMATIONAL,
                "%s disconnected: %s",
                locally_generated ? "Local" : "Remote host",msg);

  if (!common->authenticated)
    {
      /* Call the authentication_notify callback. */
      if (common->authentication_notify)
        (*common->authentication_notify)(common->user, FALSE,
                                         common->context);
    }

  /* Call the disconnect function.  Note that it is always given.
     This will call ssh_common_destroy, destroying the SshCommon object
     and all channels. */
  (*common->disconnect)(reason, locally_generated, msg, common->context);
}


/* Processes a debug message received from the connection protocol. */

void ssh_common_debug(int type, const char *msg, void *context)
{
  SshCommon common = (SshCommon)context;

  SSH_DEBUG(3, ("DEBUG received: %s", msg));

  /* Call the user-supplied debug function if non-NULL. */
  if (common->debug)
    (*common->debug)(type, msg, common->context);
}


void ssh_common_request_rekey(SshCommon common)
{




  SSH_PRECOND(common != NULL);
  SSH_PRECOND(common->chosen_host_key_alg != NULL);
  SSH_PRECOND(common->chosen_cipher_c_to_s != NULL);
  SSH_PRECOND(common->chosen_cipher_s_to_c != NULL);
  SSH_PRECOND(common->chosen_mac_c_to_s != NULL);
  SSH_PRECOND(common->chosen_mac_s_to_c != NULL);
  SSH_PRECOND(common->chosen_compression_c_to_s != NULL);
  SSH_PRECOND(common->chosen_compression_s_to_c != NULL);

  if (common->doing_rekey)
    {
      SSH_TRACE(2, ("Rekey already in progress..."));
      return;
    }

  if (*(common->compat_flags->rekey_incompatibility))
    {
      SSH_DEBUG(2,("Remote version has rekey incompatibility bug."));
      return;
    }

  common->doing_rekey = TRUE;

  SSH_TRACE(2, ("Sending rekey request..."));
  ssh_conn_send_encode(common->conn,
                       SSH_CROSS_REKEY_REQUEST,
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_cipher_c_to_s,
                       strlen(common->chosen_cipher_c_to_s),
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_cipher_s_to_c,
                       strlen(common->chosen_cipher_s_to_c),
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_mac_c_to_s,
                       strlen(common->chosen_mac_c_to_s),
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_mac_s_to_c,
                       strlen(common->chosen_mac_s_to_c),
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_compression_c_to_s,
                       strlen(common->chosen_compression_c_to_s),
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_compression_s_to_c,
                       strlen(common->chosen_compression_s_to_c),
                       SSH_FORMAT_UINT32_STR,
                       common->chosen_host_key_alg,
                       strlen(common->chosen_host_key_alg),
                       SSH_FORMAT_END);

}

void ssh_common_rekey_timeout_callback(void *context)
{
  SshCommon common = (SshCommon) context;
  SSH_PRECOND(common != NULL);

  ssh_common_request_rekey(common);
}

void ssh_common_rekey_notify_cb(void *context)
{
  SshCommon common = *(SshCommon *)context;
  SSH_PRECOND(context != NULL);

  SSH_TRACE(2, ("Rekey notification received from transport layer."));

  /* this is the case, if ssh_common_wrap() hasn't been called yet. */
  if (common == NULL)
    return;

  SSH_TRACE(2, ("Cancelling rekey timeouts..."));
  ssh_cancel_timeouts(ssh_common_rekey_timeout_callback, common);

  common->doing_rekey = TRUE;
}



/* Processes a special packet received from the connection protocol. */

void ssh_common_special(SshCrossPacketType type, const unsigned char *data,
                        size_t len, void *context)
{
  SshCommon common = (SshCommon)context;

  switch (type)
    {
    case SSH_CROSS_AUTHENTICATED:
      SSH_DEBUG(2, ("Received SSH_CROSS_AUTHENTICATED packet from "
                    "connection protocol."));
      if (ssh_decode_array(data, len,
                           SSH_FORMAT_UINT32_STR, &common->user, NULL,
                           SSH_FORMAT_END) == 0)
        ssh_fatal("ssh_common_special: bad AUTHENTICATED packet");

      common->authenticated = TRUE;

      /* Before we get here, the authentication protocol should have
         stored the authenticated user's SshUser object in
         *common->ucp (as specified in the call to
         ssh_auth_server_wrap). */
      if (common->ucp && *common->ucp)
        {
          SSH_ASSERT(common->user_data == NULL);
          common->user_data = *common->ucp;
        }
      SSH_ASSERT(common->client || common->user_data != NULL);

      /* Call the authenticated notify function if given. */
      if (common->authentication_notify)
        (*common->authentication_notify)(common->user, TRUE, common->context);

      /* Register rekey timeout. */
      if (common->config->rekey_interval_seconds > 0L)
        ssh_register_timeout(common->config->rekey_interval_seconds, 0L,
                             ssh_common_rekey_timeout_callback,
                             common);

      break;
    case SSH_CROSS_STARTUP:
      SSH_DEBUG(2, ("Received SSH_CROSS_STARTUP packet from "
                    "connection protocol."));
      /* Do nothing. */
      break;
    case SSH_CROSS_ALGORITHMS:
      common->doing_rekey = FALSE;
      ssh_xfree(common->chosen_host_key_alg);
      common->chosen_host_key_alg = NULL;
      ssh_xfree(common->chosen_cipher_c_to_s);
      common->chosen_cipher_c_to_s = NULL;
      ssh_xfree(common->chosen_cipher_s_to_c);
      common->chosen_cipher_s_to_c = NULL;
      ssh_xfree(common->chosen_mac_c_to_s);
      common->chosen_mac_c_to_s = NULL;
      ssh_xfree(common->chosen_mac_s_to_c);
      common->chosen_mac_s_to_c = NULL;
      ssh_xfree(common->chosen_compression_c_to_s);
      common->chosen_compression_c_to_s = NULL;
      ssh_xfree(common->chosen_compression_s_to_c);
      common->chosen_compression_s_to_c = NULL;

      if (ssh_decode_array(data, len,
                           /* string kex_algorithm */
                           SSH_FORMAT_UINT32_STR, NULL, NULL,
                           /* string server_host_key_algorithm */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_host_key_alg, NULL,
                           /* string server_host_key */
                           SSH_FORMAT_UINT32_STR, NULL, NULL,
                           /* string encryption_algorithm_client_to_server */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_cipher_c_to_s, NULL,
                           /* string encryption_algorithm_server_to_client */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_cipher_s_to_c, NULL,
                           /* string mac_algorithm_client_to_server */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_mac_c_to_s, NULL,
                           /* string mac_algorithm_server_to_client */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_mac_s_to_c, NULL,
                           /* string compression_algorithm_client_to_server */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_compression_c_to_s, NULL,
                           /* string compression_algorithm_server_to_client */
                           SSH_FORMAT_UINT32_STR,
                           &common->chosen_compression_s_to_c, NULL,
                           SSH_FORMAT_END) == 0)
        ssh_fatal("ssh_common_special: bad ALGORITHMS packet");

      SSH_DEBUG(2, ("Received SSH_CROSS_ALGORITHMS packet from "
                    "connection protocol."));

      if (common->authenticated)
        {
          /* Register rekey timeout. */
          if (common->config->rekey_interval_seconds > 0L)
            ssh_register_timeout(common->config->rekey_interval_seconds,
                                 0L, ssh_common_rekey_timeout_callback,
                                 common);
        }

      break;

    default:
      SSH_TRACE(0, ("Unknown special packet %d received from "
                    "connection protocol", (int)type));
      break;
    }
}


/* Creates the common processing object for the SSH server/client connection.
   This also creates the connection protocol object.
     `connection'   the connection to the other side
     `auth'         authentication protocol object
     `client'       TRUE if we are a client, FALSE if a server
     `config'       configuration data
     `peer_host_name'
                    client: name of server host
                    server: connecting client host name
     `disconnect'   function to call on disconnect (may be NULL)
     `debug'        function to call on debug message (may be NULL)
     `authenticated_notify'
                    function to call when authenticated (may be NULL)
     `destroy_notify'
                    function to call when common object is being
                    destroyed (may be NULL)
     `compat_flags' compatibility flags, stored to the common
                    struct. Should be a ssh_xmallocated object, which
                    is freed by ssh_common_destroy. Can also be NULL,
                    whence default values are used.
     `context'      context to pass to ``destroy''
   The object should be destroyed from the ``disconnect'' callback or from
   a ``close_notify'' callback (see below).  */

SshCommon ssh_common_wrap(SshStream connection,



                          SshStream auth,

                          Boolean client,
                          SshConfig config,
                          const char *peer_host_name,



                          SshConnDisconnectProc disconnect,
                          SshConnDebugProc debug,

                          SshCommonAuthenticationNotify authentication_notify,
                          SshCommonDestroyNotify destroy_notify,

                          SshCommonChannelsClosedNotify channels_closed_notify,


                          SshTransportCompat compat_flags,

                          void *context)
{
  SshCommon common;
  int i, j, num_requests, num_types;

  SSH_DEBUG(5, ("creating SshCommon object"));

  /* Sanity check: disconnect should not be NULL. */
  if (disconnect == NULL_FNPTR)
    ssh_fatal("ssh_common_wrap: disconnect must not be NULL.");

  /* Create the common object. */
  common = ssh_xcalloc(1, sizeof(*common));
  common->client = client;
  common->config = config;
  common->num_channels = 0;
  common->active_local_forwards = ssh_adt_create_generic(SSH_ADT_LIST,
                                                         SSH_ADT_DESTROY,
                                                         destroy_forward_rec,
                                                         SSH_ADT_ARGS_END);
  common->active_remote_forwards = ssh_adt_create_generic(SSH_ADT_LIST,
                                                          SSH_ADT_DESTROY,
                                                          destroy_forward_rec,
                                                          SSH_ADT_ARGS_END);
  common->active_x11_forwards = ssh_adt_create_generic(SSH_ADT_LIST,
                                                       SSH_ADT_DESTROY,
                                                       destroy_forward_rec,
                                                       SSH_ADT_ARGS_END);

  common->cleanup_hooks = ssh_adt_create_generic(SSH_ADT_LIST,
                                                 SSH_ADT_DESTROY,
                                                 ssh_adt_callback_destroy_free,
                                                 SSH_ADT_ARGS_END);



  common->disconnect = disconnect;

  common->debug = debug;
  common->auth = auth;

  common->authentication_notify = authentication_notify;
  common->destroy_notify = destroy_notify;

  common->channels_closed_notify = channels_closed_notify;

  common->context = context;
  if (common->client)
    common->server_host_name = ssh_xstrdup(peer_host_name);
  common->being_destroyed = FALSE;






  common->publickey_options = NULL;


  if (compat_flags)
    common->compat_flags = compat_flags;
  else
    common->compat_flags = ssh_xcalloc(1, sizeof(*compat_flags));


  common->last_login_time = 0;
  common->sizeof_last_login_from_host = 100;
  common->last_login_from_host =
    ssh_xcalloc(common->sizeof_last_login_from_host, sizeof(char));

  /* Get information from the connection. */
  common->remote_ip = ssh_xmalloc(100);
  if (!ssh_tcp_get_remote_address(connection, common->remote_ip, 100))
    strcpy(common->remote_ip, "UNKNOWN");
  common->remote_port = ssh_xmalloc(100);
  if (!ssh_tcp_get_remote_port(connection, common->remote_port, 100))
    strcpy(common->remote_port, "UNKNOWN");
  common->local_hostname = ssh_xmalloc(256);
  ssh_tcp_get_host_name(common->local_hostname, 256);
  common->local_ip = ssh_xmalloc(100);
  if (!ssh_tcp_get_local_address(connection, common->local_ip, 100))
    strcpy(common->local_ip, "UNKNOWN");
  common->local_port = ssh_xmalloc(100);
  if (!ssh_tcp_get_local_port(connection, common->local_port, 100))
    strcpy(common->local_port, "UNKNOWN");

  SSH_DEBUG(0, ("local ip = %s, local port = %s",
                common->local_ip, common->local_port));
  SSH_DEBUG(0, ("remote ip = %s, remote port = %s",
                common->remote_ip, common->remote_port));

  SSH_DEBUG(5, ("initializing channel types and requests"));

  /* Count the number of channel types and global requests. */
  num_types = 0;
  num_requests = 0;
  for (i = 0; ssh_channel_types[i].name; i++)
    {
      num_types++;
      for (j = 0; ssh_channel_types[i].global_requests[j].name; j++)
        num_requests++;
    }
  SSH_ASSERT(num_types == i);
































  /* Allocate memory for channel types and global requests. */
  common->global_requests = ssh_xcalloc(num_requests + 1,
                                        sizeof(common->global_requests[0]));
  common->channel_opens = ssh_xcalloc(num_types + 1,
                                      sizeof(common->channel_opens[0]));
  common->type_contexts = ssh_xcalloc(num_types,
                                      sizeof(common->type_contexts[0]));
  /* Note: the arrays are initialized to zero.  The code below relies on the
     last element (after ones that we initialize below) already being
     zeroed. */

  /* Initialize channel opens and global requests. */
  num_requests = 0;
  for (i = 0; ssh_channel_types[i].name; i++)
    {
      for (j = 0; ssh_channel_types[i].global_requests[j].name; j++)
        {
          common->global_requests[num_requests++] =
            ssh_channel_types[i].global_requests[j];
        }
      common->channel_opens[i].name = ssh_channel_types[i].name;
      common->channel_opens[i].proc = ssh_channel_types[i].open_proc;
      if (ssh_channel_types[i].type_create_proc)
        common->type_contexts[i] =
          (*ssh_channel_types[i].type_create_proc)(common);
      else
        common->type_contexts[i] = NULL;
    }

  
  /* Make sure we have something in remote_host. */
  if (!common->client)
    common->remote_host = ssh_xstrdup(peer_host_name);
  else
    common->remote_host = ssh_xstrdup(common->remote_ip);

  /* Create connection protocol object. */








  SSH_DEBUG(3, ("Creating connection protocol."));
  common->conn = ssh_conn_wrap(common->auth, SSH_CONNECTION_SERVICE,
                               common->global_requests,
                               common->channel_opens,
                               ssh_common_disconnect,
                               ssh_common_debug,
                               ssh_common_special,
                               common->compat_flags,
                               (void *)common);
  SSH_DEBUG(5, ("connection protocol created"));


  return common;
}

/* Closes all channels, destroys channel-specific and channel type specific
   data, and destroys any context common to all channels. */

void ssh_common_destroy(SshCommon common)
{
  int i;
  SshADTContainer l;
  SshADTHandle h;
  CleanUp cleanup_rec;

  /* If the common-object is already being destroyed, just return */
  if (common->being_destroyed == TRUE)
    {
      SSH_DEBUG(3, ("SshCommon already being destroyed, returning."));
      return;
    }

  SSH_DEBUG(2, ("Destroying SshCommon object."));
  
  /* We set this flag, so that we wouldn't try to destroy this struct from
     other functions or callbacks. */
  common->being_destroyed = TRUE;

  /* If there are timeouts already waiting to destroy this structure,
     cancel them*/
  ssh_cancel_timeouts(common_destroy_timeout, (void *)common);

  /* Cancel rekey timeout, as it uses SshCommon structure, which is
     now being destroyed. */
  ssh_cancel_timeouts(ssh_common_rekey_timeout_callback, (void *)common);

  /* Call cleanup hooks. */
  SSH_DEBUG(3, ("Calling clean-up hooks."));
  l = common->cleanup_hooks;
  for (h = ssh_adt_enumerate_start(l); h != SSH_ADT_INVALID;
       h = ssh_adt_enumerate_next(l, h))
    {
      cleanup_rec = ssh_adt_get(l, h);
      (*cleanup_rec->clean_up)(cleanup_rec->context);
    }
  ssh_adt_destroy(common->cleanup_hooks);
  
  /* Call channel type specific destructors and free the context array. */
  for (i = 0; ssh_channel_types[i].name; i++)
    {
      if (ssh_channel_types[i].type_destroy_proc &&
          common->type_contexts[i])
        (*ssh_channel_types[i].type_destroy_proc)(common->type_contexts[i]);
      common->type_contexts[i] = NULL;
    }

  /* Call the SshCommonDestroyNotify if given. */
  if (common->destroy_notify)
    (*common->destroy_notify)(common->context);

  ssh_xfree(common->type_contexts);
  common->type_contexts = NULL;


  /* Destroy global requests and open procs. */
  ssh_xfree(common->global_requests);
  ssh_xfree(common->channel_opens);


  /* Destroy the connection protocol context. */





  if (common->conn)
    ssh_conn_destroy(common->conn);
  common->conn = NULL;


  /* Destroy the active channel lists */
  ssh_adt_destroy(common->active_local_forwards);
  ssh_adt_destroy(common->active_remote_forwards);
  ssh_adt_destroy(common->active_x11_forwards);

  if (common->num_groups > 0)
    {
      for (i = 0; i < common->num_groups; i++)
        {
          ssh_xfree(common->group_strs[i]);
          ssh_xfree(common->gid_strs[i]);
        }
      ssh_xfree(common->group_strs);
      ssh_xfree(common->gid_strs);
      common->num_groups = 0;
    }
  /* Free allocated fields. */
  ssh_xfree(common->server_host_name);
  ssh_xfree(common->user);

  if (common->user_data)
    ssh_user_free(common->user_data, FALSE);

  ssh_xfree(common->remote_ip);
  ssh_xfree(common->remote_port);
  ssh_xfree(common->local_hostname);
  ssh_xfree(common->local_ip);
  ssh_xfree(common->local_port);
  ssh_xfree(common->compat_flags);
  ssh_xfree(common->forced_command);



  ssh_xfree(common->chosen_host_key_alg);
  ssh_xfree(common->chosen_cipher_c_to_s);
  ssh_xfree(common->chosen_cipher_s_to_c);
  ssh_xfree(common->chosen_mac_c_to_s);
  ssh_xfree(common->chosen_mac_s_to_c);
  ssh_xfree(common->chosen_compression_c_to_s);
  ssh_xfree(common->chosen_compression_s_to_c);


  if (common->remote_host)
    ssh_xfree(common->remote_host);

  if (common->last_login_from_host)
    ssh_xfree(common->last_login_from_host);















  ssh_publickey_options_free(common->publickey_options);
  common->publickey_options = NULL;

  memset(common, 'F', sizeof(*common));
  ssh_xfree(common);
}

/* Returns the channel type context for the channel type identified by
   the name. */

void *ssh_common_get_channel_type_context(SshCommon common, const char *name)
{
  int i;

  if (common->type_contexts == NULL)
    {
      SSH_DEBUG(5, ("type_contexts is already destroyed."));
      return NULL;
    }

  for (i = 0; ssh_channel_types[i].name; i++)
    {
      if (strcmp(ssh_channel_types[i].name, name) == 0)
        return common->type_contexts[i];
    }

  SSH_DEBUG(5, ("type '%s' not found", name));
  return NULL;
}

/* Informs the channel type independent code that a channel has been
   destroyed.  This may destroy the SshCommon object if there are no
   more channels, causing a call to the channel type specific destroy
   function.  Care should be taken to call this function as the last thing
   done in a channel destroy function. */

void ssh_common_destroy_channel(SshCommon common)
{
  common->num_channels--;
  SSH_DEBUG(1, ("num_channels now %d", common->num_channels));


  if (common->num_channels == 0 && !common->being_destroyed)
    {
      /* The following may or may not, cause the destruction of
         the common struct). */
      if (common->channels_closed_notify)
        (*common->channels_closed_notify)(common->context);
    }
















}

/* Informs the channel type independent code that a new channel has been
   created. */

void ssh_common_new_channel(SshCommon common)
{
  common->num_channels++;
  SSH_DEBUG(1, ("num_channels now %d", common->num_channels));
}

/* Adds a new channel to the beginning of the active list. */

void ssh_common_add_active_forward(SshTcpForwardActive fwd_active)
{
  SshADTContainer list = NULL;

  switch(fwd_active->channel_type)
    {
    case SSH_TFACT_LOCAL:
      list = fwd_active->common->active_local_forwards;
      break;
    case SSH_TFACT_REMOTE:
      list = fwd_active->common->active_remote_forwards;
      break;
    case SSH_TFACT_X11:
      list = fwd_active->common->active_x11_forwards;
      break;
    }

  SSH_VERIFY(ssh_adt_insert_to(list, SSH_ADT_END, (void*)fwd_active) !=
             SSH_ADT_INVALID);
}


/* Deletes the destroyed channel from the active channel list */
void ssh_common_delete_active_forward(void* context)
{
  SshTcpForwardActive delete_node = (SshTcpForwardActive)context;
  SshCommon common;
  SshADTContainer list = NULL;

  if (context == NULL) return;

  common = delete_node->common;

  switch(delete_node->channel_type)
    {
    case SSH_TFACT_LOCAL:
      list = common->active_local_forwards;
      break;
    case SSH_TFACT_REMOTE:
      list = common->active_remote_forwards;
      break;
    case SSH_TFACT_X11:
      list = common->active_x11_forwards;
      break;
    }

  /* find the node in the list and delete it */

  if (ssh_adt_detach_object(list, context) == SSH_ADT_INVALID)
    {
      SSH_TRACE(2, ("Forward not found in list (bug?)."));
    }

  ssh_common_destroy_active_forward(delete_node);

  /* tell the common object that the channel is destroyed */
  ssh_common_destroy_channel(common);
}

void ssh_publickey_options_free(SshAuthPublicKeyOptions pubkey_options)
{
  if (pubkey_options == NULL)
    return;

  if (pubkey_options->use_allowed_from)
    ssh_adt_destroy(pubkey_options->use_allowed_from);
  if (pubkey_options->use_denied_from)
    ssh_adt_destroy(pubkey_options->use_denied_from);

  if (pubkey_options->environment_vars)
    ssh_adt_destroy(pubkey_options->environment_vars);

  ssh_xfree(pubkey_options);
}

void ssh_common_register_clean_up(SshCommon common,
                                  SshCommonCleanUpCB clean_up, void *context)
{
  CleanUp cleanup_rec;
  SshADTHandle h;
  SSH_PRECOND(common->cleanup_hooks != NULL);

  cleanup_rec = ssh_xmalloc(sizeof(*cleanup_rec));
  cleanup_rec->clean_up = clean_up;
  cleanup_rec->context = context;
  h = ssh_adt_insert(common->cleanup_hooks, cleanup_rec);
  SSH_VERIFY(h != SSH_ADT_INVALID);
}
